#
# Revision: 3.0.180532
#
# Copyright (c) 2018 Dell Technologies Inc.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# DELL EMC CORPORATION. USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF DELL EMC.
#
package SCAgent;
use lib qw(. ./sc);
use strict;
use warnings;
use parent 'ISerializable';
use Crypt::CBC;
use ag_i18n_inc;
use sc::api;
use SCTools qw(RUN RUN_SINGLE MIN MAX ISIN FILTER UNIQ GREPUNIQ PRINTDUMP TOSTRING SAFEPARAMS);
use SCLogger;
use SCAgentState;
use SCAgentVPair;
use SCAgentServer;



my ($SINGLE_INSTANCE, %PROTOTYPE, %AGENT_STABLE_STATES, %AGENT_GOING_ONLINE, %AGENT_GOING_OFFLINE);
my ($OS, %CMD_FCHBA, %CMD_RESCAN, %CMD_RESCAN_SOFTLY);



use constant {

    SC_ACTION_ARGS          =>  'ACTION_ARGS',
    SC_ATTR_DSM_PRIMARY     =>  'DSMPrimary',
    SC_ATTR_DSM_STANDBY     =>  'DSMStandby',
    SC_ATTR_DSM_PORT        =>  'DSMPort',
    SC_ATTR_USERNAME        =>  'UserName',
    SC_ATTR_PASSWORD        =>  'Password',
    SC_ATTR_VOLUMES         =>  'Volumes',
    SC_ATTR_SERVERS         =>  'Servers',
    SC_ATTR_LOGFILE         =>  'LogFile',
    SC_ATTR_AUTOTAKEOVER    =>  'AutoTakeover',
    SC_ATTR_STATE           =>  'State',
    SC_ATTR_ISTATE          =>  'IState',
    SC_VCS_RESOURCEINFO     =>  'ResourceInfo',
    SC_VCS_AEPTIMEOUT       =>  'AEPTimeout',
    SC_VCS_TIMEOUT_MONITOR  =>  'MonitorTimeout',
    SC_VCS_TIMEOUT_ONLINE   =>  'OnlineTimeout',
    SC_VCS_TIMEOUT_INFO     =>  'InfoTimeout',
    SC_VCS_RESOURCE_NAME    =>  'ResourceName',
    SC_VCS_CONFIG_READONLY  =>  'VCS.ReadOnly',
    SC_API_CONNECTION       =>  'API.Connection',
    SC_START_TIME           =>  'StartTime',
    SC_IS_CONSISTENT        =>  'IsConsistent'
};



use constant {

    HARES_STATE_ONLINE      => 'ONLINE',
    HARES_TYPE_DISKGROUP    => 'DiskGroup',
    HARES_ATTR_STATE        => 'State',
    HARES_ATTR_TYPE         => 'Type',
    HARES_ATTR_GROUP        => 'Group',
    HARES_ATTR_DISKGROUP    => 'DiskGroup'
};



use constant {

    MONITOR_OFFLINE     => 100,
    MONITOR_ONLINE      => 110,
    MONITOR_ACCEPTABLE  => 101,
    MONITOR_UNKNOWN     =>  99
};



use constant {

    TIMEOUT_MONITOR_MIN     => 120,
    TIMEOUT_MONITOR_PER_DEV => 5,
    TIMEOUT_ONLINE_MIN      => 900,
    TIMEOUT_ONLINE_PER_DEV  => 60
};



use constant {

    ACTION_CONFIG_MODE_INIT     => 'CONFIG_MODE_INIT',
    ACTION_CONFIG_MODE_TOPOLOGY => 'CONFIG_MODE_TOPOLOGY',
    ACTION_CONFIG_MODE_VOLUMES  => 'CONFIG_MODE_VOLUMES'
};



use constant {

    #ONLINE
    #OFFLINE
    #FAULTED
    #OFFLINE|MONITOR TIMEDOUT
    #OFFLINE|STATE UNKNOWN
    #OFFLINE|ADMIN WAIT
    #ONLINE|RESTARTING
    #ONLINE|MONITOR TIMEDOUT
    #ONLINE|STATE UNKNOWN
    #ONLINE|UNABLE TO OFFLINE
    #ONLINE|ADMIN WAIT
    #FAULTED|MONITOR TIMEDOUT
    #FAULTED|STATE UNKNOWN

    AGENT_STATE_OFFLINE                     => 0,
    AGENT_STATE_ONLINE                      => 2,

    AGENT_ISTATE_NOTWAITING                 => 0,   #Resource is not in transition.
    AGENT_ISTATE_GOING_ONLINE               => 1,   #WAITING TO GO ONLINE - Agent notified to bring the resource online but procedure not yet complete.
    AGENT_ISTATE_GOING_ONLINE_CHILD         => 2,   #WAITING FOR CHILDREN ONLINE - Resource to be brought online, but resource depends on at least one offline resource. Resource transitions to waiting to go online when all children are online.
    AGENT_ISTATE_GOING_OFFLINE              => 3,   #WAITING TO GO OFFLINE - Agent notified to take the resource offline but procedure not yet complete.
    AGENT_ISTATE_GOING_OFFLINE_PROPAGATE    => 4,   #WAITING TO GO OFFLINE (propagate) - Same as above, but when completed the resource's children will also be offline.
    AGENT_ISTATE_GOING_ONLINE_REVERSE       => 5,   #WAITING TO GO ONLINE (reverse) - Resource waiting to be brought online, but when it is online it attempts to go offline. Typically this is the result of issuing an offline command while resource was waiting to go online.
    AGENT_ISTATE_GOING_OFFLINE_PATH         => 6,   #WAITING TO GO OFFLINE (path) - Agent notified to take the resource offline but procedure not yet complete. When the procedure completes, the resource's children which are a member of the path in the dependency tree will also be offline.
    AGENT_ISTATE_GOING_OFFLINE_REVERSE      => 7,   #WAITING TO GO OFFLINE (reverse) - Resource waiting to be brought offline, but when it is offline it attempts to go online. Typically this is the result of issuing an online command while resource was waiting to go offline.
    AGENT_ISTATE_GOING_ONLINE_REVERSEPATH   => 8,   #WAITING TO GO ONLINE (reverse/path) - Resource waiting to be brought online, but when online it is brought offline. Resource transitions to WAITING TO GO OFFLINE (path). Typically this is the result of fault of a child resource while resource was waiting to go online.
    AGENT_ISTATE_GOING_OFFLINE_PARENT       => 9,   #WAITING FOR PARENT OFFLINE;
};



use constant {

    RESULT_OK       => 'OK',
    RESULT_SKIP     => 'SKIP(OK)',
    RESULT_FAIL     => 'FAILED',
    RESULT_UNKNOWN  => 'UNKNOWN'
};



use constant {

    OS_AIX      => 'AIX',
    OS_HPUX     => 'HPUX',
    OS_LINUX    => 'LINUX',
    OS_SOLARIS  => 'SOLARIS',
    OS_WINDOWS  => 'WINDOWS'
};



use constant {

    JOB_ONLINE  => 'online',
    JOB_MAPPING => 'mapping',
    JOB_REVERSE => 'reverse'
};







BEGIN {

    $SINGLE_INSTANCE = 0;

    $OS = OS_AIX()      if $^O =~ /aix/i;
    $OS = OS_HPUX()     if $^O =~ /hpux/i;
    $OS = OS_LINUX()    if $^O =~ /linux/i;
    $OS = OS_SOLARIS()  if $^O =~ /solaris/i;
    $OS = OS_WINDOWS()  if $^O =~ /MSWin/i;
    $OS = OS_LINUX()    unless $OS;


    %CMD_FCHBA = (

        OS_AIX()        => 'for i in $(lscfg | grep fcs | awk "{print $2}"); do echo $i && lscfg -vpl $i; done',
        OS_HPUX()       => 'for fc in $(ioscan -fnkC fc | grep "/dev" | awk \'{print $1}\'); do fcmsutil $fc; done',
        OS_LINUX()      => 'cat /sys/class/fc_host/*/port_name',
        OS_SOLARIS()    => 'fcinfo hba-port',
        OS_WINDOWS()    => 'echo windows OS is not supported at the moment'
    );



    %CMD_RESCAN = (

        OS_AIX()        => ['cfgmgr -v', 'vxdctl enable', 'vxdisk list', 'vxdisk -x LUN_SERIAL_NO -x PRIV_UDID -p list'],
        OS_HPUX()       => ['ioscan -fnC disk', 'insf -C disk', 'vxdctl enable', 'vxdisk list', 'vxdisk -x LUN_SERIAL_NO -x PRIV_UDID -p list'],
        OS_LINUX()      => ['if [ -f /usr/bin/rescan-scsi-bus.sh ]; then /usr/bin/rescan-scsi-bus.sh; else for i in $(ls /sys/class/scsi_host/); do echo "- - -" > /sys/class/scsi_host/$i/scan; done; fi', 'vxdctl enable', 'vxdisk list', 'vxdisk -x LUN_SERIAL_NO -x PRIV_UDID -p list'],
        OS_SOLARIS()    => ['devfsadm -Cv', 'vxdctl enable', 'vxdisk scandisks', 'vxdisk list'],
        OS_WINDOWS()    => ['echo "rescan" | diskpart', 'vxdctl enable', 'vxdisk scandisks', 'vxdisk list']
    );



    %CMD_RESCAN_SOFTLY = (

        OS_AIX()        => ['vxdctl -f enable', 'vxdisk list', 'vxdisk -x LUN_SERIAL_NO -x PRIV_UDID -p list'],
        OS_HPUX()       => ['vxdctl -f enable', 'vxdisk list', 'vxdisk -x LUN_SERIAL_NO -x PRIV_UDID -p list'],
        OS_LINUX()      => ['vxdctl -f enable', 'vxdisk list', 'vxdisk -x LUN_SERIAL_NO -x PRIV_UDID -p list'],
        OS_SOLARIS()    => ['vxdctl -f enable', 'vxdisk list', 'vxdisk -x LUN_SERIAL_NO -x PRIV_UDID -p list'],
        OS_WINDOWS()    => ['vxdctl -f enable', 'vxdisk list', 'vxdisk -x LUN_SERIAL_NO -x PRIV_UDID -p list']
    );



    %AGENT_STABLE_STATES = (
        AGENT_ISTATE_NOTWAITING()               => 1,
        AGENT_ISTATE_GOING_ONLINE_CHILD()       => 1,
        AGENT_ISTATE_GOING_OFFLINE_PARENT()     => 1
    );



    %AGENT_GOING_ONLINE = (
        AGENT_ISTATE_GOING_ONLINE()             => 1,
        AGENT_ISTATE_GOING_ONLINE_REVERSE()     => 1,
        AGENT_ISTATE_GOING_ONLINE_REVERSEPATH() => 1
    );



    %AGENT_GOING_OFFLINE = (
        AGENT_ISTATE_GOING_OFFLINE()            => 1,
        AGENT_ISTATE_GOING_OFFLINE_PROPAGATE()  => 1,
        AGENT_ISTATE_GOING_OFFLINE_PATH()       => 1,
        AGENT_ISTATE_GOING_OFFLINE_REVERSE()    => 1
    );



    %PROTOTYPE = (

        SC_ATTR_DSM_PRIMARY()   =>  '',
        SC_ATTR_DSM_STANDBY()   =>  '',
        SC_ATTR_DSM_PORT()      =>  '',
        SC_ATTR_USERNAME()      =>  '',
        SC_ATTR_PASSWORD()      =>  '',
        SC_ATTR_VOLUMES()       =>  {},
        SC_ATTR_SERVERS()       =>  {},
        SC_ATTR_LOGFILE()       =>  '',
        SC_ATTR_AUTOTAKEOVER()  =>   0,
        SC_ATTR_STATE()         =>  -1,
        SC_ATTR_ISTATE()        =>  -1,
        SC_VCS_AEPTIMEOUT()	    =>   0,
        SC_VCS_RESOURCE_NAME()  =>  '',
        SC_VCS_CONFIG_READONLY()=>   0,
        SC_API_CONNECTION()     =>   0,
        SC_START_TIME()         =>   0,
        SC_IS_CONSISTENT()      =>   1,
        SC_ACTION_ARGS()        =>  []
    );

}





sub new {

    return $SINGLE_INSTANCE if $SINGLE_INSTANCE;

    my  $class = shift;
    my  $self  = bless( { %PROTOTYPE }, (ref($class) || $class) );
        $self->{SC_VCS_RESOURCE_NAME()} = shift || shift(@ARGV) || 'UnknownSCAgent';
        $self->{SC_IS_CONSISTENT()} = 0 unless $self->{SC_VCS_RESOURCE_NAME()};
        $self->{SC_START_TIME()} = time();

    foreach    (SC_ATTR_DSM_PRIMARY(),
                SC_ATTR_DSM_STANDBY(),
                SC_ATTR_DSM_PORT(),
                SC_ATTR_USERNAME(),
                SC_ATTR_PASSWORD(),
                SC_ATTR_VOLUMES(),
                SC_ATTR_SERVERS(),
                SC_ATTR_LOGFILE(),
                SC_ATTR_AUTOTAKEOVER(),
                SC_ATTR_STATE(),
                SC_ATTR_ISTATE(),
                SC_VCS_AEPTIMEOUT()) {

        $self->{SC_IS_CONSISTENT()} &&= ARGUMENT($_, \$self->{$_});
    }

    SCLogger::addLogFiles($self->{SC_ATTR_LOGFILE()} =
       ($self->{SC_ATTR_LOGFILE()} || '/tmp/sc.agent').'.'.($self->{SC_VCS_RESOURCE_NAME()} || 'unknown'));

    if ($self->{SC_IS_CONSISTENT()}) {

        my %vols = %{$self->{SC_ATTR_VOLUMES()}||{}};
        my %srvs = %{$self->{SC_ATTR_SERVERS()}||{}};
        $self->{SC_ATTR_VOLUMES()} = [ map { SCAgentVPair ->_decode($self, $_, $vols{$_}) } keys %vols ];
        $self->{SC_ATTR_SERVERS()} = [ map { SCAgentServer->_decode($self, $_, $srvs{$_}) } keys %srvs ];

        TRACE MSG(MSG_AGENT_INIT_SUCCEED), $self;
    }

    else {

        TRACE "AGENT INIT FAILED", {%$self};
        TRACE "AGENT ARGV", [@ARGV];
        LOG    MSG_AGENT_INIT_FAILED(), @ARGV;
    }

    return $SINGLE_INSTANCE = $self;
}





sub VXDISK_UPDATE { RUN([], 'vxdisk -Ccf updateudid', @_) }

sub HARES_MODIFY  { RUN([], $ag_i18n_inc::HARES, '-modify', SAFEPARAMS(@_) ) }

sub HARES_LOCAL   { RUN([], $ag_i18n_inc::HARES, '-local',  SAFEPARAMS(@_) ) }

sub HARES_ACTION  { RUN([], $ag_i18n_inc::HARES, '-action', SAFEPARAMS(@_) ) }

sub HACONF_DUMP   { RUN([], $ag_i18n_inc::HACONF, (shift() ? '-dump -makero' : '-dump')) }

sub HACONF_MAKERW { RUN([], $ag_i18n_inc::HACONF, '-makerw') }

sub HACLUS_RONLY  { RUN_SINGLE(($_[0] ||  1), $ag_i18n_inc::HACLUS, '-value', 'ReadOnly') }

sub HASYS_NODEID  { RUN_SINGLE(($_[0] || ''), $ag_i18n_inc::HASYS,  '-nodeid',    @_) }

sub HARES_VALUE   { RUN_SINGLE(($_[2] || ''), $ag_i18n_inc::HARES,  '-value',     $_[0], $_[1]) }

sub HARES_STATE   { RUN_SINGLE(''           , $ag_i18n_inc::HARES,  '-state',     @_) }

sub HARES_LINK    { RUN_SINGLE(''           , $ag_i18n_inc::HARES,  '-link',      @_) }

sub HARES_FLUSH   { RUN_SINGLE(''           , $ag_i18n_inc::HARES,  '-flushinfo', @_ > 1 ? @_ : (@_, '-localclus')) }

sub HARES_OVERRIDE{ RUN_SINGLE(''           , $ag_i18n_inc::HARES,  '-override',  @_) }

sub HARES_OVALUES { my @result = (); return () unless RUN(\@result, '(', $ag_i18n_inc::HARES, '-display -ovalues', @_, ') || echo'); return @result }

# hares -dep Compellent
##Group       Parent     Child
#testSG       testDG     Compellent
#testSG       testSG2    Compellent
sub HARES_DEP     { my @result = (); return () unless RUN(\@result, $ag_i18n_inc::HARES, '-dep', @_); return () unless @result > 1; shift @result; return grep { $_ } map { (split /\s+/)[1] } @result }

sub HASYS_LIST    { my @result = (); return () unless RUN(\@result, $ag_i18n_inc::HASYS, '-list',           SAFEPARAMS(@_)); return @result }

sub HAGRP_RESRS   { my @result = (); return () unless RUN(\@result, $ag_i18n_inc::HAGRP, '-resources',      SAFEPARAMS(@_)); return @result }

sub VXDMP_CTLR    { my @result = (); return () unless RUN(\@result, 'vxdmpadm getctlr all');    return WWNS(@result) }

#vxddladm list hbas
#HBA-ID  DRIVER          FIRMWARE        DISCOVERY                 STATE    ADDRESS
#======================================================================================
#c15     -               -               -                         Online  -
#c1      qla2xxx         8.03.01 (d0d5)  fabric                    Online  20:00:00:0e:1e:c2:c0:50
#c4      qla2xxx         7.03.00 (d0d5)  fabric                    Online  20:00:00:0e:1e:c2:c0:71
#c0      -               -               -                         Online  -
#c2      qla2xxx         8.03.01 (d0d5)  fabric                    Online  20:00:00:0e:1e:c2:c0:51
#c3      qla2xxx         7.03.00 (d0d5)  fabric                    Online  20:00:00:0e:1e:c2:c0:70
#c16     -               -               -                         Online  -
sub VXDDL_HBA     { my @result = (); return () unless RUN(\@result, 'vxddladm list hbas');      return WWNS(@result) }

sub VXDDL_PORTS   { my @result = (); return () unless RUN(\@result, 'vxddladm list ports');     return WWNS(@result) }

sub VXDISK_LIST   { my @result = (); return () unless RUN(\@result, 'vxdisk -q', @_ ? ('-g', SAFEPARAMS(@_)) : (), 'list'); return @result }

sub VXDISK_LUNSN  { my @result = (); return () unless RUN(\@result, 'vxdisk -q', @_ ? ('-g', SAFEPARAMS(@_)) : (), '-x LUN_SERIAL_NO -p list'); return @result }

sub VXDISKS_MAP   { map { my @d = split /\s+/; (@d > 1) ? ($d[1] => $d[0]) : () } VXDISK_LUNSN(@_) }

sub RESCAN        { map { my @result = (); RUN(\@result, $_) ? @result : "Failed: ($_)" } grep { $_ } @{$CMD_RESCAN{$OS}||[]} };

sub RESCAN_SOFTLY { map { my @result = (); RUN(\@result, $_) ? @result : "Failed: ($_)" } grep { $_ } @{$CMD_RESCAN_SOFTLY{$OS}||[]} };

sub THIS_NODE     { my $node = HASYS_NODEID; my @candidates = grep { index($node, $_) >= 0 } HASYS_LIST; return $candidates[0] if 1 == @candidates; return $1 if HASYS_NODEID =~ /Node name:\s+(\w+)\s+Node ID:/i; return '' }

sub WWNS          { grep { length($_) == 16 } map { s/://rg } grep { $_ } map { $1 if /\b(?:0x)?(([0-9A-F]{2}:?){8})\b/gi } @_ }

sub MATCHWWN      { my ($wwn) = WWNS(shift); return 0 unless $wwn;
                    my  @wwn  = map { hex } unpack('A2'x8, $wwn);
                    map { my $x = 0; return 1 if 1 >= scalar( grep { $wwn[$x++] != hex($_) } unpack('A2'x8, $_) ) } @_;
                    return 0;
                  }

sub ARGUMENT      {

    my ($argument, $output) = @_;
    my  $result;

    while (ref($output) eq 'REF') { $output = ${$output} };

    if (ref($output) eq 'SCALAR') {

       ($result, ${$output}) = VCSAG_GET_ATTR_VALUE($argument, -1, 1, @ARGV);
        return 1 if $result == $VCSAG_SUCCESS;
        return LOG(MSG_ARGUMENT_FAILED(), $argument);
    }

    elsif ((ref($output) eq 'ARRAY') || (ref($output) eq 'HASH')) {

        my ($size, $index, $item, @temp);
       ($result, $size, $index)= VCSAG_GET_ATTR_VALUE($argument, @ARGV);
        return LOG(MSG_ARGUMENT_FAILED, $argument) unless $result == $VCSAG_SUCCESS;

        for (my $i = 1; $i <= $size; $i++) {

           ($result, $item) = VCSAG_GET_ATTR_VALUE($argument, $index, $i, @ARGV);
            return LOG(MSG_ARGUMENT_FAILED(), $argument, "i = $i", "size = $size", "index = $index") unless $result == $VCSAG_SUCCESS;
            push @temp, $item;
        }

        @{$output} = @temp if ref($output) eq 'ARRAY';
        %{$output} = @temp if ref($output) eq 'HASH';
        return 1;
    }

    return LOG(MSG_ARGUMENT_FAILED(), $argument, 'SCAgent::ARGUMENT sub requires $output param to be a reference to SCALAR, ARRAY or HASH');
}





#@method
#@override
sub properties { qw (name volumes servers timeout timeHeadroom isStable isOnline isOffline isValid autotakeover) }


#@method
#@returns sc::api
sub connect         {

    my $self = shift;
    return 0 unless $self->isValid;
    return $self->{SC_API_CONNECTION()} if $self->{SC_API_CONNECTION()};

    my $cipher = Crypt::CBC->new(-key => $self->{SC_ATTR_USERNAME()}, -cipher => 'OpenSSL::AES');
    my $passwd = $self->{SC_ATTR_PASSWORD()};
       $passwd = $cipher->decrypt_hex($passwd) if $passwd =~ /\A(?:0?[xX])?(?:_?[0-9a-fA-F])*\z/;

    my sc::api $api = sc::api->connect (sc::api::DSMServer()        => $self->{SC_ATTR_DSM_PRIMARY()},
                                        sc::api::DSMServerReplica() => $self->{SC_ATTR_DSM_STANDBY()},
                                        sc::api::DSMPort()          => $self->{SC_ATTR_DSM_PORT()},
                                        sc::api::DSMUser()          => $self->{SC_ATTR_USERNAME()},
                                        sc::api::DSMPass()          => $passwd);

    if ($self->{SC_API_CONNECTION()} = $api) {

        if ($passwd eq $self->{SC_ATTR_PASSWORD()}) {

            $self->vxConfigOpen;
            $self->attrSet(SC_ATTR_PASSWORD(), $cipher->encrypt_hex($self->{SC_ATTR_PASSWORD()}));
            $self->vxConfigClose;
        }

        LOG MSG_CONNECTION_STANDBY if $api->isFailedOver;
        return $self->{SC_API_CONNECTION()};
    }

    else {

        LOG MSG_CONNECTION_FAILED;
        LOG MSG_CONNECTION_NOSTDBY unless $self->{SC_ATTR_DSM_STANDBY()};
    }

    return 0;
}


#@method
#@returns sc::api
sub api             { return $_[0]->{SC_API_CONNECTION()} || $_[0]->connect }


#@method
sub disconnect      { return $_[0]->api->disconnect if $_[0]->{SC_API_CONNECTION()} }


#@method
sub localNodes      { return UNIQ( map { $_->ssn } shift->servers() ) }


#@method
sub volumes         { return @{shift->{SC_ATTR_VOLUMES()}||[]} }


#@method
sub servers         { my ($self, @ssn) = @_; return @{$self->{SC_ATTR_SERVERS()}||[]} unless @ssn; return FILTER($self->{SC_ATTR_SERVERS()}, 'ssn', @ssn) }


#@method
sub name            { return shift->{SC_VCS_RESOURCE_NAME()} ||'' }


#@method
sub autotakeover    { return shift->{SC_ATTR_AUTOTAKEOVER()} || 0 }


#@method
sub isValid         { return shift->{SC_IS_CONSISTENT()} || 0 }


#@method
sub timeout         { return shift->{SC_VCS_AEPTIMEOUT()} || 60 }


#@method
sub timeStart       { return shift->{SC_START_TIME()} }


#@method
sub timeDuration    { return MAX(0, time - $_[0]->timeStart) }


#@method
sub timeLeft        { return MAX(0, $_[0]->timeStart + $_[0]->timeout - $_[0]->timeHeadroom - time) }


#@method
sub timeHeadroom    { return MIN(30, $_[0]->timeout * 0.1) }


#@method
sub isStable        { return exists($AGENT_STABLE_STATES{shift->{SC_ATTR_ISTATE()}}) }


#@method
sub isGoingOnline   { return exists($AGENT_GOING_ONLINE {shift->{SC_ATTR_ISTATE()}}) }


#@method
sub isGoingOffline  { return exists($AGENT_GOING_OFFLINE{shift->{SC_ATTR_ISTATE()}}) }


#@method
sub isOnline        { return AGENT_STATE_ONLINE()  == shift->{SC_ATTR_STATE()} }


#@method
sub isOffline       { return AGENT_STATE_OFFLINE() == shift->{SC_ATTR_STATE()} }


#@method
sub isAttachedTo    { my ($self, $ssn) = @_; return ISIN($ssn, map { $_->ssn } $self->servers()) }


#@method
sub isCurrentHost   { my ($self, $id ) = @_; return ISIN($id,  map { $_->id  } $self->servers()) }


#@method
sub isRescanNeeded  { my ($self, $vxdisks) = ($_[0], {VXDISKS_MAP()}); return scalar(grep { $_->isRescanRecommended($vxdisks) } $self->volumes) }


#@method
sub attribute       { my ($self, $attribute, $default) = @_; return HARES_VALUE($self->name, $attribute, ($default || '')) }


#@method
sub attrSet         { my ($self, $attr, @values) = @_;; HARES_MODIFY($self->name, $attr, @values) }


#@method
sub attrClear       { my ($self, $attr) = @_; HARES_MODIFY($self->name, $attr, '-delete', '-keys') }


#@method
sub attrDeleteKey   { my ($self, $attr, $key) = @_; HARES_MODIFY($self->name, $attr, '-delete', $key) }


#@method
sub attrAddKey      { my ($self, $attr, $key, @values) = @_; HARES_MODIFY($self->name, $attr, '-add', $key, @values) ||  HARES_MODIFY($self->name, $attr, '-update', $key, @values) }


#@method
sub attrUpdateKey   { my ($self, $attr, $key, @values) = @_; HARES_MODIFY($self->name, $attr, '-update', $key, @values) || HARES_MODIFY($self->name, $attr, '-add', $key, @values) }


#@method
sub attrOverride    { my ($self, $attr, @values) = @_; HARES_OVERRIDE($self->name, $attr) unless (grep {/\b$attr\b/} (HARES_OVALUES($self->name))); $self->attrSet($attr, @values) }


#@method
sub attrLocalize    { my $self = shift; HARES_LOCAL($self->name, @_) }


#@method
sub attrFlushInfo   { my $self = shift; HARES_FLUSH($self->name, @_) }


#@method
sub state           { my ($self, $sys) = @_; return $sys ? HARES_STATE($self->name, '-sys', $sys) : HARES_STATE($self->name, '-sys', THIS_NODE) }


#@method
sub actionArgs      { my $self = shift; unless (@{$self->{SC_ACTION_ARGS()}||[]}) { ARGUMENT(SC_ACTION_ARGS(), \$self->{SC_ACTION_ARGS()}) }; return @{ $self->{SC_ACTION_ARGS()} || [] } }


#@method
sub vxConfigOpen    { HACONF_MAKERW if shift->{SC_VCS_CONFIG_READONLY()} = HACLUS_RONLY }


#@method
sub vxConfigClose   { HACONF_DUMP(shift->{SC_VCS_CONFIG_READONLY()}) }


#@method
sub vxGroup         { HARES_VALUE(shift->{SC_VCS_RESOURCE_NAME()}, HARES_ATTR_GROUP()) }


#@method
sub vxResources     { grep { HARES_ATTR_DISKGROUP() eq HARES_VALUE($_, HARES_ATTR_TYPE()) } HAGRP_RESRS(shift->vxGroup) }


#@method
sub vxParents       { return grep { HARES_TYPE_DISKGROUP() eq HARES_VALUE($_, HARES_ATTR_TYPE()) } HARES_DEP(shift->name) }


#@method
sub vxDiskGroups    { grep { $_ } map { HARES_VALUE($_, HARES_ATTR_DISKGROUP()) } shift->vxParents }


#@method
sub vxDiskNames     { grep { $_ } map { $1 if /^(\w+)[\s\b]/ } map { VXDISK_LUNSN($_) } shift->vxDiskGroups }


#@method
sub vxDiskSerials   { grep { $_ } map { $1 if /\b([0-9A-F]{8}-[0-9A-F]{8})\b/i } map { VXDISK_LUNSN($_) } shift->vxDiskGroups }


#@method
sub vxDisks         { map { VXDISKS_MAP($_) } shift->vxDiskGroups }





#@method
sub areParentsOnline {

    my $self = shift;
    my @parents = $self->vxParents;
    return 0 unless scalar(@parents);
    return 1 if    (scalar(@parents) == scalar(grep { HARES_STATE_ONLINE() eq HARES_VALUE($_, HARES_ATTR_STATE()) } @parents));
}





#@method
sub vxRescan {

    my  $self = shift;
    my  $t = $self->timeLeft; LOG(MSG_AGENT_RESCAN_START(),  $t, '[full rescan]');
    my  @r = RESCAN;
        $t-= $self->timeLeft; LOG(MSG_AGENT_RESCAN_FINISH(), $t, @r);

    return @r;
}





#@method
sub vxRescanSoftly {

    my  $self = shift;
    my  $t = $self->timeLeft; LOG(MSG_AGENT_RESCAN_START(),  $t,'[rescan softly]');
    my  @r = RESCAN_SOFTLY;
        $t-= $self->timeLeft; LOG(MSG_AGENT_RESCAN_FINISH(), $t, @r);

    return @r;
}





#@method
sub prepareVxDisks {

    my  $self   = shift;
    my  %vxdisk = (VXDISKS_MAP);
    if (%vxdisk) {

        my @invisible = grep {my SCAgentVPair $vp = $_; not scalar( grep {exists($vxdisk{$_})} $vp->serials ) } $self->volumes;

        if (@invisible) {

            $self->vxRescan;
            $self->doWait(30);
            %vxdisk = (VXDISKS_MAP);
            @invisible = grep {my SCAgentVPair $vp = $_; not scalar( grep {exists($vxdisk{$_})} $vp->serials ) } $self->volumes;
        }

        unless (@invisible) {

            my @visible = grep {exists($vxdisk{$_})} map {my SCAgentVPair $vp = $_; $vp->serials} $self->volumes;

            VXDISK_UPDATE @vxdisk{@visible};
        }
    }
}





#@method
sub doWait {

    my  $self = shift;
    my  $wait = shift || 10;
    my  $time = $self->timeLeft;

        $wait = MIN($wait, $time);

    TRACE(MSG(MSG_AGENT_SLEEP), {'sleep for' => $wait, 'time left' => $time}) if $wait > 0;
          LOG(MSG_AGENT_TIMEOUT) unless $time > 0;

    sleep($wait) if $wait > 0;
}





#@method
sub updateVpair {

    my $self = shift;
    my SCAgentVPair $vp = shift;

            $self->vxConfigOpen;
    my $r = $self->attrUpdateKey(SC_ATTR_VOLUMES(), $vp->key, $vp->_encode);
            $self->vxConfigClose;

    return $r;
}





#@method
sub configure {

    my  $self = shift;
    my  $mode = ($self->actionArgs)[0] || ACTION_CONFIG_MODE_INIT();
    my  $result = 1;
    LOG(MSG_CONFIG(), $mode);

    return LOG(MSG_AGENT_DISCONNECTED) unless $self->connect;



    if ($mode eq ACTION_CONFIG_MODE_INIT()) {

        LOG MSG_CONFIG_INIT();

        unless ($self->vxParents) {

            my @dg = grep { not HARES_LINK($_, $self->name) } $self->vxResources;
            LOG(MSG_CONFIG_ERR_LINK(), @dg) if @dg;
        }

        if ($self->vxParents) {

            $self->vxConfigOpen;
            $result = _runAcrossNodes_ConfigTopology($self);
            $result = _runAcrossNodes_ConfigVolumes ($self) && $result; #keep in mind perls lazy boolean operations if we want the agent to run both Topology and Volumes configurations
            $self->vxConfigClose;

            $result ? LOG(MSG_CONFIG_SUCCESS()) : LOG(MSG_CONFIG_FAILURE());
            $result ? exit 0 : exit 1;
        }

        else {

            LOG MSG_CONFIG_ERR_PARENTS();
            exit 2;
        }
    }



    elsif ($mode eq ACTION_CONFIG_MODE_TOPOLOGY()) {

        LOG(MSG_CONFIG_TOPOLOGY());

        $result = $self->_configTopology;
        $result ? LOG(MSG_CONFIG_TOPOLOGY_SUCCESS()) : LOG(MSG_CONFIG_TOPOLOGY_FAILURE());
        $result ? exit 0 : exit 1;
    }



    elsif ($mode eq ACTION_CONFIG_MODE_VOLUMES()) {

        LOG(MSG_CONFIG_VOLUMES());

        $result = $self->_configVolumes;
        $result ? LOG(MSG_CONFIG_VOLUMES_SUCCESS()) : LOG(MSG_CONFIG_VOLUMES_FAILURE());
        $result ? exit 0 : exit 1;
    }
}





#@method
sub _runAcrossNodes_ConfigTopology {

    my  $self       = shift;
    my  $me         = THIS_NODE;
    my  %systems    = (map { $_ => $self->state($_) } HASYS_LIST); #grep { $_ ne 'sfha2' }
    my  %result     = (map { $_ => RESULT_UNKNOWN() } keys %systems);
    my  $err        = 0;

            LOG(MSG_CONFIG_TOPOLOGY());
    return  LOG(MSG_CONFIG_ERR_NODELIST())  unless %systems;
    return  LOG(MSG_CONFIG_ERR_NODEID())    unless $me;
    return  LOG(MSG_CONFIG_ERR_NODEID())    unless $systems{$me};
            LOG(MSG_CONFIG_ERR_CLEAR(), SC_ATTR_SERVERS()) unless $self->attrClear(SC_ATTR_SERVERS());

    foreach (keys %systems) {

        my  $r = ($_ eq $me) ? $self->_configTopology : HARES_ACTION($self->name, 'configure', '-actionargs', ACTION_CONFIG_MODE_TOPOLOGY(), '-sys', $_);
            $result{$_} = $r ? RESULT_OK() : RESULT_FAIL();
            $err += 1 unless $r;

        LOG(MSG_CONFIG_TOPOLOGY_FAILURE(), $_) unless $r;
    }

    PRINTDUMP "TOPOLOGY CONFIG RESULT", {%result};
    TRACE     "TOPOLOGY CONFIG RESULT", {%result};
    return 0 == $err;
}





#@method
sub _runAcrossNodes_ConfigVolumes {

    my  $self       = shift;
    my  $me         = THIS_NODE;
    my  @systems    = HASYS_LIST;
    my  %result     = (map { $_ => RESULT_SKIP() } @systems);
    my  $err = 1;

            LOG(MSG_CONFIG_VOLUMES());
    return  LOG(MSG_CONFIG_ERR_NODELIST())  unless @systems;
    return  LOG(MSG_CONFIG_ERR_NODEID())    unless $me;
    return  LOG(MSG_CONFIG_ERR_NODEID())    unless $result{$me};

    foreach (@systems) {

        my  $r = ($_ eq $me) ? $self->_configVolumes : HARES_ACTION($self->name, 'configure', '-actionargs', ACTION_CONFIG_MODE_VOLUMES(), '-sys', $_);
            $result{$_} = $r ? RESULT_OK() : RESULT_FAIL();
            $err = 0 if $r;
            last if $r;
    }

    LOG(MSG_CONFIG_VOLUMES_FAILURE(), $_) if $err;
    PRINTDUMP "VOLUMES CONFIG RESULT", {%result};
    TRACE     "VOLUMES CONFIG RESULT", {%result};
    return 0 == $err;
}





#@method
sub _configTopology {

    my  $self = shift;
    my  @vxPorts = VXDDL_HBA;
        TRACE "LOCAL HBA WWNs", [@vxPorts];
        return LOG(MSG_CONFIG_ERR_VXPORTS()) unless @vxPorts;

    my  @scPorts = $self->api->hbas;
        TRACE "SC PORTS", [@scPorts];
        return LOG(MSG_CONFIG_ERR_SCPORTS()) unless @scPorts;

    my  @path = grep { my sc::hba $h = $_; scalar (grep { MATCHWWN($_, @vxPorts) } $h->wwns) } grep { $_->serverID } grep { $_->isValid } @scPorts;
        TRACE "DETECTED PATHS", [@path], qw(name server ssn);
        return LOG(MSG_CONFIG_ERR_NOPATH()) unless @path;

    my  @srvs = map { my sc::hba $h = $_; SCAgentServer->new($self, $h->server) } GREPUNIQ('serverID', @path);
    my  %srvs = map { my SCAgentServer $s = $_; $s->key => $s->_encode } @srvs;

    my  $code = 1;

        $code &&= LOG(MSG_CONFIG_ATTR_FAILURE()) unless $self->attrLocalize(SC_ATTR_SERVERS());
        $code &&= LOG(MSG_CONFIG_ATTR_FAILURE()) unless $self->attrClear(SC_ATTR_SERVERS());
        $code &&= LOG(MSG_CONFIG_ATTR_FAILURE()) if grep { !$self->attrAddKey(SC_ATTR_SERVERS(), $_, $srvs{$_}) } keys %srvs;
        $self->{SC_ATTR_SERVERS()} = [@srvs];
        TRACE "LOCAL SERVERS",       [@srvs];

    return $code;
}





#@method
sub _configVolumes {

    my  $self = shift;
        return LOG(MSG_CONFIG_ERR_DISKGROUPS()) unless $self->areParentsOnline;

    my  @devs = $self->vxDiskSerials;
        return LOG(MSG_CONFIG_ERR_SERIALS()) unless @devs;
        TRACE "DISK GROUPS",[$self->vxDiskGroups];
        TRACE "DEVICES",    [$self->vxDiskNames];
        TRACE "SERIALS",    [@devs];

    my  @vols = GREPUNIQ('serial', $self->api->volumes(@devs)); #live volumes produces two volumes (prim + sec) with the identical serial numbers
        return LOG(MSG_CONFIG_ERR_LUNS()) unless @vols;
        TRACE "VOLUMES",    [@vols];

    my  %miss = map  { my sc::volume $v = $_; $v->serial => 1 } @vols;
    my  @miss = grep { not exists($miss{$_}) } @devs;
        TRACE "MISSING VOLUMES", [@miss] if @miss;
        return LOG(MSG_CONFIG_ERR_LUNS(), @miss) if @miss;

    my  $rpts = [$self->api->restorePoints];
    my  @rpts = ();
        push @rpts, FILTER($rpts, 'sourceVolumeID',    map { $_->id } @vols );
        push @rpts, FILTER($rpts, 'targetVolumeID',    map { $_->id } @vols );
        push @rpts, FILTER($rpts, 'activatedVolumeID', map { $_->id } @vols );
        TRACE "RESTORE POINTS", [@rpts];

    my  %rpts = map  { my sc::restorepoint $r = $_; ($r->sourceVolumeID => $r, $r->targetVolumeID => $r, $r->isDRActivated ? ($r->activatedVolumeID => $r) : ()) } @rpts;
        @miss = grep { my sc::volume $v = $_; not exists($rpts{$v->id}) } @vols;
        TRACE "MISSING RESTORE POINTS", [@miss] if @miss;
        return LOG(MSG_CONFIG_ERR_RPTS,  @miss) if @miss;

    my  $dset = [ map { my sc::volume $v = $_; SCAgentVPair->new($self, $rpts{$v->id}) } @vols ];
        $self->{SC_ATTR_VOLUMES()} = $dset;
        TRACE "DATASET", $dset;

    my  $code = 1;
        $code &&= LOG(MSG_CONFIG_ATTR_FAILURE()) unless $self->attrClear(SC_ATTR_VOLUMES());

    foreach my SCAgentVPair $vp (@$dset) {

        $code &&= LOG(MSG_CONFIG_ATTR_FAILURE()) unless $self->attrAddKey(SC_ATTR_VOLUMES(), $vp->key, $vp->_encode);
    }

    if ($code) {

        LOG(MSG_CONFIG_ATTR_INFO_VOLS(), map { $_->key } @$dset);

        my $mto = MAX(TIMEOUT_MONITOR_MIN(), TIMEOUT_MONITOR_PER_DEV() * scalar(@$dset));
        my $oto = MAX(TIMEOUT_ONLINE_MIN(),  TIMEOUT_ONLINE_PER_DEV()  * scalar(@$dset));
        TRACE "REQUIRED MINIMAL TIMEOUTS", {monitor => $mto, online => $oto, info => $mto};

        $self->attrOverride(SC_VCS_TIMEOUT_MONITOR, $mto) if scalar($self->attribute(SC_VCS_TIMEOUT_MONITOR) || 0) < $mto;
        $self->attrOverride(SC_VCS_TIMEOUT_ONLINE,  $oto) if scalar($self->attribute(SC_VCS_TIMEOUT_ONLINE)  || 0) < $oto;
        $self->attrOverride(SC_VCS_TIMEOUT_INFO,    $mto) if scalar($self->attribute(SC_VCS_TIMEOUT_INFO)    || 0) < $mto;
        $self->attrFlushInfo;
        map { $self->attrAddKey(SC_VCS_RESOURCEINFO, $_->key, TOSTRING($_->serials)) } @$dset;
    }

    return $code;
}





#@method
sub runVolumeJob {

    my  $self   = shift;
    my  $job    = shift || '(job is not specified)';
    my  $pause  = shift || 30;
    my  $cycle  = 0;
    my  $title = '[SC AGENT][JOB = ' . $job . '] ';
    my  %dataset = map { $_->prepareNewJob; $_->key => $_ } $self->volumes;
    my  @ready  = ();

    unless (scalar(%dataset)) {

        TRACE $title, 'dataset is empty', [$self->volumes];
        return 0;
    }

    unless ($job && SCAgentVPair->can($job)) {

        TRACE $title, ('Method (' . $job . ') is not implemented in the SCAgentVPair class');
        return 0;
    }

    TRACE $title . 'Dataset', [$self->volumes];

    while (%dataset && $self->timeLeft) {

        TRACE   $title.'[CYCLE = '.(++$cycle).'] [TIME LEFT: '.($self->timeLeft).']', { %dataset };
        my      @goodtogo = (grep {$dataset{$_}->$job} keys %dataset);
        TRACE   $title.'[CYCLE = '.($cycle).'] finish successfully in this round', [ @dataset{@goodtogo} ], SCAgentVPair->properties(), 'history' if @goodtogo;
        push    @ready, @dataset{@goodtogo} if @goodtogo;
        delete  @dataset{@goodtogo} if @goodtogo;
        TRACE   $title.'[CYCLE = '.($cycle).'] still waiting', [ values %dataset ], SCAgentVPair->properties(), 'history' if %dataset;

                $self->vxRescanSoftly   if %dataset && $job eq JOB_MAPPING() && $cycle < 3;
                $self->vxRescan         if %dataset && $job eq JOB_MAPPING() && $cycle > 2;
                $self->doWait($pause)   if %dataset;
    }

    TRACE $title.'result', [@ready, values %dataset], SCAgentVPair->properties(), 'history';

    return 0 == %dataset;
}





#@method
sub GoOnline {

    my $self   = shift;
    my %result = ();

   #$self->vxRescan if $self->isRescanNeeded;
    $self->vxRescan;
    $result{JOB_ONLINE()} = $self->runVolumeJob(JOB_ONLINE()) ? 'OK' : 'FAILED';

   #$self->vxRescanSoftly if $self->isRescanNeeded;
    $self->vxRescanSoftly;
    $result{JOB_MAPPING()} = $self->runVolumeJob(JOB_MAPPING()) ? 'OK' : 'FAILED';
    $result{JOB_REVERSE()} = 'SKIPPED';

    unless (SCAgentState->new($self)->isRpointHealthy) {

        LOG(MSG_ONLINE_INCOMPLETE);
    }

    else {

        $self->prepareVxDisks;
        $result{JOB_REVERSE()} = $self->runVolumeJob(JOB_REVERSE()) ? 'OK' : 'FAILED';
        $self->api->validateAll;

        if (my @unreversed = grep { $_->swap } $self->volumes) {

            LOG(MSG_ONLINE_NEEDSWAP, @unreversed);
        }

        else {

            LOG(MSG_ONLINE_SUCCESS, $self->volumes);
        }
    }

    TRACE '[ONLINE][RESULT] vxdisk map',    { VXDISKS_MAP() };
    TRACE '[ONLINE][RESULT] vxdisk list',   [ VXDISK_LIST() ];
    TRACE '[ONLINE][RESULT] jobs',          { %result };
    TRACE '[ONLINE][RESULT] timings',       { duration => $self->timeDuration, 'time left' => $self->timeLeft };
}





1;