#
# Revision: 3.0.180532
#
# Copyright (c) 2018 Dell Technologies Inc.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# DELL EMC CORPORATION. USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF DELL EMC.
#
package SCAgentServer;
use lib qw(. ./sc);
use strict;
use warnings;
use parent 'ISerializable';
use SCTools qw(ISIN SAFEVALUE SAFEKEY);
use SCLogger;


use constant {

    SERVER_ID         => 'id',
    SERVER_SSN        => 'ssn',
    SERVER_Name       => 'name',
    SERVER_Key        => 'key',
    SERVER_Agent      => 'agent'
};





sub new {

    my              $class  = shift;
    my              $agent  = shift;
    my sc::server   $server = shift;
                    $server = $server->cluster;

    return bless {

            SERVER_Agent()   => $agent,
            SERVER_ID()      => $server->id,
            SERVER_SSN()     => $server->ssn,
            SERVER_Name()    => SAFEVALUE($server->name),

        }, $class;
}





sub _decode {

    my ($class, $agent, $key, $body) = @_;
    my  $proto = JSON::from_json $body;
        $proto->{SERVER_Agent()} = $agent;
        $proto->{SERVER_Key()}   = $key;

    return bless $proto, $class;
}





sub _encode {

    my $self = shift;
    return JSON::to_json( { map { $_ => $self->{$_} } (SERVER_ID, SERVER_SSN, SERVER_Name) } );
}





#@method
#@override
sub properties  { qw (name ssn) }


#@method
#@returns SCAgent
sub agent       { $_[0]->{SERVER_Agent()} }


#@method
#@returns sc::api
sub api         { $_[0]->agent->api }



#@method
sub id          { $_[0]->{SERVER_ID()}    || 0  }



#@method
sub ssn         { $_[0]->{SERVER_SSN()}   || 0  }



#@method
sub name        { $_[0]->{SERVER_Name()}  || '' }



#@method
sub key         { $_[0]->id ? SAFEKEY($_[0]->name) . ' (' . $_[0]->id . ')' : '' }



#@method
sub isLocal     { ISIN($_[0]->ssn, $_[0]->agent->localNodes) }



#@method
sub isHealthy {

    my $self = shift;
    return 1 if $self->api->_submit('GET', sc::api::URLServer(), $self->id)->asObject('sc::server')->isValid;
    return not $self->api->get_sc($self->ssn)->isUp
}



1;