#
# Revision: 3.0.180532
#
# Copyright (c) 2018 Dell Technologies Inc.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# DELL EMC CORPORATION. USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF DELL EMC.
#
package SCLogger;
use strict;
use warnings;
use ag_i18n_inc;
use SCTools qw(DUMP ISOTIME);
use POSIX qw(strftime);
use constant {

    MSG_VCS_UNHEALTHY_REPLICA   =>  118,
    MSG_VCS_ERROR_ATTRIBUTES    =>  128,

    MSG_API_REQUEST_FAILED      =>  705,
    MSG_RUN_FAILED              =>  905,
    MSG_ARGUMENT_FAILED         =>  908,
    MSG_DEBUG                   =>  915,

    MSG_AGENT_INIT_SUCCEED      => 1001,
    MSG_AGENT_INIT_FAILED       => 1009,
    MSG_AGENT_INIT_ATTR_ERR     => 1008,
    MSG_AGENT_DISCONNECTED      => 1019,
    MSG_AGENT_SLEEP             => 1020,
    MSG_AGENT_TIMEOUT           => 1025,
    MSG_AGENT_RESCAN_START      => 1030,
    MSG_AGENT_RESCAN_FINISH     => 1031,
    MSG_AGENT_MISCONFIGURED     => 1049,
    MSG_AGENT_UNHEALTHY_SERVERS => 1059,
    MSG_AGENT_UNHEALTHY_REPLICA => 1069,
    MSG_AGENT_MIXED_DIRECTIONS  => 1075,

    MSG_CONNECTION_SUCCEED      => 1201,
    MSG_CONNECTION_FAILED       => 1209,
    MSG_CONNECTION_PRIMARY      => 1211,
    MSG_CONNECTION_PRIMARY_ERR  => 1219,
    MSG_CONNECTION_FAILOVER     => 1221,
    MSG_CONNECTION_FAILOVER_ERR => 1229,
    MSG_CONNECTION_STANDBY      => 1215,
    MSG_CONNECTION_NOSTDBY      => 1225,

    MSG_MONITOR_EXIT_CODE       => 2100,
    MSG_MONITOR_UNKNOWN         => 2109,
    MSG_MONITOR_STABLE          => 2110,
    MSG_MONITOR_GOING_ONLINE    => 2120,
    MSG_MONITOR_GOING_OFFLINE   => 2130,
    MSG_MONITOR_CONFIG_EMPTY    => 2141,

    MSG_ONLINE_SUCCESS          => 2201,
    MSG_ONLINE_NEEDSWAP         => 2205,
    MSG_ONLINE_INCOMPLETE       => 2208,
    MSG_ONLINE_GOODTOGO         => 2301,
    MSG_ONLINE_ERROR            => 2308,
    MSG_ONLINE_UNKNOWN          => 2309,
    MSG_ONLINE_WAIT             => 2310,
    MSG_ONLINE_WAIT_LV_SWAP     => 2320,
    MSG_ONLINE_WAIT_DRACTIVATE  => 2330,
    MSG_ONLINE_WAIT_TARGET      => 2340,
    MSG_ONLINE_WAIT_MIRRBACK    => 2350,
    MSG_ONLINE_WAIT_SWAP        => 2360,
    MSG_ONLINE_WAIT_RESTORE     => 2370,
    MSG_ONLINE_GO_AWAY_SOURCE   => 2385,
    MSG_ONLINE_GO_AWAY_TARGET   => 2395,
    MSG_ONLINE_GO_AWAY_INVISIBLE=> 2398,

    MSG_VPAIR_UNKNOWN           => 3008,
    MSG_VPAIR_RUNNINGGOOD       => 3010,
    MSG_VPAIR_SOURCEFAILED      => 3020,
    MSG_VPAIR_DRACTIVATING      => 3030,
    MSG_VPAIR_DRACTIVATED       => 3040,
    MSG_VPAIR_RESTORING         => 3050,
    MSG_VPAIR_NEEDDEACTIVATE    => 3060,
    MSG_VPAIR_RETRY             => 3070,
    MSG_VPAIR_RETRY_LIMIT       => 3075,
    MSG_VPAIR_TASK_RUNNING      => 3100,
    MSG_VPAIR_TASK_COMPLETE     => 3110,
    MSG_VPAIR_TASK_FAILED       => 3125,
    MSG_VPAIR_SWAP_STARTED      => 3200,
    MSG_VPAIR_SWAP_CANCELED     => 3205,
    MSG_VPAIR_SWAP_READY        => 3210,
    MSG_VPAIR_SWAP_WAIT         => 3220,
    MSG_VPAIR_SWAP_FAILED       => 3235,
    MSG_VPAIR_RPDEF_SUCCESS     => 3300,
    MSG_VPAIR_RPDEF_ERR_INVALID => 3318,
    MSG_VPAIR_RPDEF_ERR_STATE   => 3325,
    MSG_VPAIR_RPDEF_ERR_SERVER  => 3338,
    MSG_VPAIR_RPDEF_ERR_FAILED  => 3348,
    MSG_VPAIR_VXDISK_MISSING    => 3358,

    MSG_CONFIG                  => 4000,
    MSG_CONFIG_SUCCESS          => 4001,
    MSG_CONFIG_FAILURE          => 4009,
    MSG_CONFIG_INIT             => 4010,
    MSG_CONFIG_TOPOLOGY         => 4020,
    MSG_CONFIG_TOPOLOGY_SUCCESS => 4021,
    MSG_CONFIG_TOPOLOGY_FAILURE => 4028,
    MSG_CONFIG_VOLUMES          => 4030,
    MSG_CONFIG_VOLUMES_SUCCESS  => 4031,
    MSG_CONFIG_VOLUMES_FAILURE  => 4038,
    MSG_CONFIG_ATTR_SUCCESS     => 4041,
    MSG_CONFIG_ATTR_FAILURE     => 4049,
    MSG_CONFIG_ATTR_INFO_VOLS   => 4051,
    MSG_CONFIG_ATTR_INFO_SSNS   => 4061,
    MSG_CONFIG_ATTR_INFO_SRVS   => 4071,
    MSG_CONFIG_ERR_PARENTS      => 4108,
    MSG_CONFIG_ERR_NODELIST     => 4118,
    MSG_CONFIG_ERR_NODEID       => 4128,
    MSG_CONFIG_ERR_CLEAR        => 4138,
    MSG_CONFIG_ERR_VXPORTS      => 4148,
    MSG_CONFIG_ERR_SCPORTS      => 4158,
    MSG_CONFIG_ERR_NOPATH       => 4168,
    MSG_CONFIG_ERR_DISKGROUPS   => 4175,
    MSG_CONFIG_ERR_VXDISKS      => 4188,
    MSG_CONFIG_ERR_SERIALS      => 4198,
    MSG_CONFIG_ERR_LUNS         => 4208,
    MSG_CONFIG_ERR_RPTS         => 4218,
    MSG_CONFIG_ERR_RPDEFINE     => 4228,
    MSG_CONFIG_ERR_LINK         => 4238
};







my (%MESSAGES, %MESSAGELEVELS, $LOGFILES);



BEGIN {

    %MESSAGELEVELS = (

        0 => 'I',
        1 => 'N',
        5 => 'W',
        8 => 'E',
        9 => 'C',
    );

    %MESSAGES = (

        MSG_VCS_ERROR_ATTRIBUTES()      => '[SC-AGENT]',

        MSG_API_REQUEST_FAILED()        => '[SC-API] request failed',
        MSG_RUN_FAILED()				=> '[SC-AGENT] Failure to RUN command',
        MSG_ARGUMENT_FAILED()			=> '[SC-AGENT] Failure to get VCSAgent argument',
        MSG_DEBUG()						=> '[DEBUG]',

        MSG_AGENT_INIT_SUCCEED()		=> '[SC-AGENT] has been successfully initialized',
        MSG_AGENT_INIT_FAILED()			=> '[SC-AGENT] initialization has failed',
        MSG_AGENT_INIT_ATTR_ERR()		=> '[SC-AGENT] failure to get SCAgent attribute',
        MSG_AGENT_DISCONNECTED()        => '[SC-AGENT] SC Agent is disconnected from Dell Storage Manager',
        MSG_AGENT_SLEEP()				=> '[SC-AGENT] SC Agent sleeps',
        MSG_AGENT_TIMEOUT()				=> '[SC-AGENT] time is out. Exiting soon',
        MSG_AGENT_RESCAN_START()		=> '[SC-AGENT] RESCAN has started. Time left',
        MSG_AGENT_RESCAN_FINISH()		=> '[SC-AGENT] RESCAN is finished. Duration',
        MSG_AGENT_MISCONFIGURED()       => '[SC-AGENT] SC Agent configuration is invalid. Please, check: (1) agent parameters, (2) SC Restore Point availability and status, (3) SC Volumes availability, (4) SC Nodes availability and (5) DSM Server availability. In case of storage or replication configuration was changed outside of SC Agent control, please reconfigure the SC Agent by running configure action on the node where all DiskGroups are in ONLINE state',
        MSG_AGENT_UNHEALTHY_SERVERS()   => '[SC-AGENT] SC Servers, configured with this Agent, are not valid. Please update configuration',
        MSG_AGENT_UNHEALTHY_REPLICA()   => '[SC-AGENT] SC Volumes, controlled by this SC Agent, are in UNHEALTHY state. Switchover and Failover are not possible when SC Volumes are in Unprotected state. Please check storage configuration and especially replication states and restore points status',
        MSG_AGENT_MIXED_DIRECTIONS()    => '[SC-AGENT] SC Replications run in opposite directions (some volumes are local for this node, the other are remote). Please check replications and reverse them if needed in order to restore consistency of the volume set',

        MSG_CONNECTION_PRIMARY()        => '[SC-API] Connection to SC REST API is established (DSM PRIMARY)',
        MSG_CONNECTION_PRIMARY_ERR()    => '[SC-API] failure to connect to primary SC REST API endpoint',
        MSG_CONNECTION_FAILOVER()       => '[SC-API] Connection to SC REST API is successfully failovered to DSM STANDBY',
        MSG_CONNECTION_FAILOVER_ERR()   => '[SC-API] failure to connect to secondary SC REST API endpoint',
        MSG_CONNECTION_SUCCEED()		=> '[SC-API] Connection to SC REST API endpoint has been successfully established',
        MSG_CONNECTION_FAILED()			=> '[SC-API] Failure to connect to SC REST API',
        MSG_CONNECTION_STANDBY()		=> '[SC-API] SC REST API failed to connect to DSM Primary server and is successfully failovered to DSM Standby (Replica) server',
        MSG_CONNECTION_NOSTDBY()		=> '[SC-API] DSM Standby (Replica) server has not been setup',

        MSG_MONITOR_STABLE()			=> '[SC-MONITOR] SCAgent State is stable',
        MSG_MONITOR_UNKNOWN()			=> '[SC-MONITOR] SCAgent State is unknown',
        MSG_MONITOR_GOING_ONLINE()		=> '[SC-MONITOR] SCAgent is going Online',
        MSG_MONITOR_GOING_OFFLINE()		=> '[SC-MONITOR] SCAgent is going Offline',
        MSG_MONITOR_CONFIG_EMPTY()      => '[SC-MONITOR] SCAgent is not configured. Will let it go ONLINE in order to allow DiskGroups to come ONLINE and thus to let the agent to be configured later on',
        MSG_MONITOR_EXIT_CODE()			=> '[SC-MONITOR] Monitor exit code',

        MSG_ONLINE_SUCCESS()            => '[SC-ONLINE] SUCCESS: all devices under replication are write enabled and visible on the host',
        MSG_ONLINE_NEEDSWAP()           => '[SC-ONLINE] INCOMPLETE RESULT: failure to reverse some of the volume replications, though the DR-activated copies of those volumes are write enabled and visible to the host, so the host is good to go. Monitor function will report an Unknown State until those replicas are reversed and brought back to a normal state',
        MSG_ONLINE_INCOMPLETE()         => '[SC-ONLINE] INCOMPLETE RESULT: some Volume Pairs are not visible on the host. Probably, they just need more time to complete replication failover or swap. Please, consult the agent logs for more details',
        MSG_ONLINE_GOODTOGO()			=> '[SC-ONLINE] decision = GOOD-TO-GO',
        MSG_ONLINE_ERROR()              => '[SC-ONLINE] ERROR = RestorePoint is broken and is in error state. Admin intervention is needed',
        MSG_ONLINE_UNKNOWN()		    => '[SC-ONLINE] ERROR = unknown state. Admin intervention is needed',
        MSG_ONLINE_WAIT()		        => '[SC-ONLINE] decision = WAIT (need more time for the current task to complete)',
        MSG_ONLINE_WAIT_LV_SWAP()       => '[SC-ONLINE] decision = LIVEVOLUME SWAP ROLES',
        MSG_ONLINE_WAIT_DRACTIVATE()    => '[SC-ONLINE] decision = ACTIVATE DR',
        MSG_ONLINE_WAIT_TARGET()        => '[SC-ONLINE] decision = WAIT for DrActivated Volume to come up',
        MSG_ONLINE_WAIT_MIRRBACK()      => '[SC-ONLINE] decision = MIRROR BACK',
        MSG_ONLINE_WAIT_SWAP()          => '[SC-ONLINE] decision = SWAP VPair and reconfigure the SCAgent',
        MSG_ONLINE_WAIT_RESTORE()       => '[SC-ONLINE] decision = RESTORE',
        MSG_ONLINE_GO_AWAY_SOURCE()     => '[SC-ONLINE] WRONG SITE: you probably should not be here, cause this is a Source Site and the Source Volume is failed. Go away to a remote site or another node where Target Volume could be activated and become accessible',
        MSG_ONLINE_GO_AWAY_TARGET()     => '[SC-ONLINE] WRONG SITE: you probably should not be here, cause the Source Volume isn not available, and its Target Volume pair is activated on a remote site. Go away to that site',
        MSG_ONLINE_GO_AWAY_INVISIBLE()  => '[SC-ONLINE] WRONG SITE: you probably should not be here, cause neither Primary nor Secondary volumes can be accessed from this node',

        MSG_VPAIR_UNKNOWN()				=> '[SC-ONLINE] VPAIR STATE = RP State is unknown',
        MSG_VPAIR_RUNNINGGOOD()			=> '[SC-ONLINE] VPAIR STATE = Running Good',
        MSG_VPAIR_SOURCEFAILED()		=> '[SC-ONLINE] VPAIR STATE = Source Volume has failed',
        MSG_VPAIR_DRACTIVATING()		=> '[SC-ONLINE] VPAIR STATE = Activating DR Site',
        MSG_VPAIR_DRACTIVATED()			=> '[SC-ONLINE] VPAIR STATE = DR Site has been activated',
        MSG_VPAIR_RESTORING()			=> '[SC-ONLINE] VPAIR STATE = Restoring back to Source',
        MSG_VPAIR_NEEDDEACTIVATE()		=> '[SC-ONLINE] VPAIR STATE = Need to deactivate DR-volume in order to complete the Restore',
        MSG_VPAIR_RETRY()               => '[SC-ONLINE] RETRYING the last operation',
        MSG_VPAIR_RETRY_LIMIT()         => '[SC-ONLINE] RETRY attempts hit the limit. Wont retry anymore',
        MSG_VPAIR_TASK_RUNNING()		=> '[SC-ONLINE] ASYNC TASK is still running',
        MSG_VPAIR_TASK_COMPLETE()		=> '[SC-ONLINE] ASYNC TASK has completed',
        MSG_VPAIR_TASK_FAILED()         => '[SC-ONLINE] ASYNC TASK has failed',
        MSG_VPAIR_SWAP_STARTED()        => '[SC-ONLINE] VPAIR Volume SWAP is registered',
        MSG_VPAIR_SWAP_CANCELED()       => '[SC-ONLINE] VPAIR Volume SWAP is canceled',
        MSG_VPAIR_SWAP_READY()          => '[SC-ONLINE] VPAIR Volume SWAP: new RestorePoint is ready. Going to save the new configuration',
        MSG_VPAIR_SWAP_WAIT()           => '[SC-ONLINE] VPAIR Volume SWAP: still waiting for new replication pair is ready and RecoverPoint is saved',
        MSG_VPAIR_SWAP_FAILED()         => '[SC-ONLINE] VPAIR Volume SWAP: failure to save new config. Will try again during the next cycle',
        MSG_VPAIR_RPDEF_SUCCESS()       => '[SC-PREDEFINE] RestorePoint has been successfully predefined',
        MSG_VPAIR_RPDEF_ERR_INVALID()	=> '[SC-PREDEFINE] RestorePoint is invalid',
        MSG_VPAIR_RPDEF_ERR_STATE()		=> '[SC-PREDEFINE] RestorePoint is not in the RunningGood state. It is likely that PredefineDR procedure would not work for it. But we will try',
        MSG_VPAIR_RPDEF_ERR_SERVER()	=> '[SC-PREDEFINE] failure to get server list for target SSN. Please, run *configure* action (in Cluster Manager GUI -> right-click menu -> Actions) to setup Topology configuration correctly',
        MSG_VPAIR_RPDEF_ERR_FAILED()	=> '[SC-PREDEFINE] the PredefineDR procedure has failed for RestorePoint',
        MSG_VPAIR_VXDISK_MISSING()      => '[SC-VPAIR][VXDISK] volume Serial Number is not visible on the host',

        MSG_CONFIG()					=> '[SC-CONFIG] configuration process is started',
        MSG_CONFIG_SUCCESS()			=> '[SC-CONFIG] configuration has completed successfully',
        MSG_CONFIG_FAILURE()			=> '[SC-CONFIG] one or more nodes failed to complete configuration. This operation is successful only when all nodes of the VCS are available and configured successfully',
        MSG_CONFIG_INIT()				=> '[SC-CONFIG][INIT] starting config cycle through all the nodes of the cluster',
        MSG_CONFIG_TOPOLOGY()			=> '[SC-CONFIG][TOPOLOGY] going to update TOPOLOGY configuration',
        MSG_CONFIG_TOPOLOGY_SUCCESS()	=> '[SC-CONFIG][TOPOLOGY] configuration has been successfully updated',
        MSG_CONFIG_TOPOLOGY_FAILURE()	=> '[SC-CONFIG][TOPOLOGY] failure to update configuration',
        MSG_CONFIG_VOLUMES()			=> '[SC-CONFIG][VOLUMES] going to update VOLUMES configuration',
        MSG_CONFIG_VOLUMES_SUCCESS()	=> '[SC-CONFIG][VOLUMES] configuration has been successfully updated',
        MSG_CONFIG_VOLUMES_FAILURE()	=> '[SC-CONFIG][VOLUMES] failure to update configuration',
        MSG_CONFIG_ATTR_SUCCESS()		=> '[SC-CONFIG][ATTRIBUTE] SCAgent settings are successfully updated',
        MSG_CONFIG_ATTR_FAILURE()		=> '[SC-CONFIG][ATTRIBUTE] failure to update SCAgent settings',
        MSG_CONFIG_ATTR_INFO_VOLS()		=> '[SC-CONFIG] Volumes attribute is successfully updated',
        MSG_CONFIG_ATTR_INFO_SRVS()		=> '[SC-CONFIG] Servers attribute is successfully updated',
        MSG_CONFIG_ATTR_INFO_SSNS()		=> '[SC-CONFIG] LocalNodes attribute is successfully updated',
        MSG_CONFIG_ERR_PARENTS()		=> '[SC-CONFIG][CHECK] failure to get DiskGroup parents linked to this SC Agent. In order to configure an SC Agent it should be linked to at least one DiskGroup parent resource. All parents linked to an SC Agent should be brought ONLINE state before configuring the SC Agent',
        MSG_CONFIG_ERR_NODELIST()		=> '[SC-CONFIG][CHECK] failure to list nodes of the Veritas Cluster',
        MSG_CONFIG_ERR_NODEID()			=> '[SC-CONFIG][CHECK] failure to determine the local VCS node name',
        MSG_CONFIG_ERR_CLEAR()			=> '[SC-CONFIG][CHECK] can not clear SC Agent attribute',
        MSG_CONFIG_ERR_VXPORTS()		=> '[SC-CONFIG][TOPOLOGY] failure to get local host WWNs. Does the host contains FC adapters?',
        MSG_CONFIG_ERR_SCPORTS()		=> '[SC-CONFIG][TOPOLOGY] failure to get a list of host HBAs from SC-nodes',
        MSG_CONFIG_ERR_NOPATH()			=> '[SC-CONFIG][TOPOLOGY] failure to trace path from SC-arrays to this host. Probably, the host isn\'t visible from the host side. Is Server objects are setup properly?',
        MSG_CONFIG_ERR_DISKGROUPS()		=> '[SC-CONFIG][VOLUMES] failure to get DiskGroup parent resources: they are not linked to the Agent or not all of them are online. The Agent should be linked to DiskGroup resources and all of them should be in the ONLINE state',
        MSG_CONFIG_ERR_VXDISKS()		=> '[SC-CONFIG][VOLUMES] failure to get VXdisk(s) of the DiskGroup',
        MSG_CONFIG_ERR_SERIALS()		=> '[SC-CONFIG][VOLUMES] failure to get Serial Number for VxDisks',
        MSG_CONFIG_ERR_LUNS()			=> '[SC-CONFIG][VOLUMES] failure to get SC Volumes corresponding to VxDisk serial numbers',
        MSG_CONFIG_ERR_RPTS()			=> '[SC-CONFIG][VOLUMES] failure to get SC Restore Points for corresponding Volumes',
        MSG_CONFIG_ERR_RPDEFINE()		=> '[SC-CONFIG][VOLUMES] failure to predefine restore point',
        MSG_CONFIG_ERR_LINK()           => '[SC-CONFIG][CHECK] failure to link DiskGroup parents to the SC Agent resource. Please be aware that in order to configure SC Agent properly you need to link it to appropriate DiskGroup resources'
    );



    unless ($LOGFILES) {

        VCSAG_SET_ENVS($ARGV[0] || 'SCAgent', 3355362);
        $LOGFILES = [];
    }
}



BEGIN {

    require Exporter;
    our @ISA = qw(Exporter);
    our @EXPORT = qw(
                    LOG
                    MSG
                    TRACE
                    MSG_DEBUG
                    MSG_RUN_FAILED
                    MSG_ARGUMENT_FAILED
                    MSG_AGENT_INIT_SUCCEED
                    MSG_AGENT_INIT_FAILED
                    MSG_AGENT_INIT_ATTR_ERR
                    MSG_AGENT_DISCONNECTED
                    MSG_AGENT_SLEEP
                    MSG_AGENT_TIMEOUT
                    MSG_AGENT_RESCAN_START
                    MSG_AGENT_RESCAN_FINISH
                    MSG_AGENT_MISCONFIGURED
                    MSG_AGENT_UNHEALTHY_SERVERS
                    MSG_AGENT_UNHEALTHY_REPLICA
                    MSG_AGENT_MIXED_DIRECTIONS
                    MSG_CONNECTION_PRIMARY
                    MSG_CONNECTION_PRIMARY_ERR
                    MSG_CONNECTION_FAILOVER
                    MSG_CONNECTION_FAILOVER_ERR
                    MSG_CONNECTION_SUCCEED
                    MSG_CONNECTION_FAILED
                    MSG_CONNECTION_STANDBY
                    MSG_CONNECTION_NOSTDBY
                    MSG_MONITOR_EXIT_CODE
                    MSG_MONITOR_UNKNOWN
                    MSG_MONITOR_STABLE
                    MSG_MONITOR_GOING_ONLINE
                    MSG_MONITOR_GOING_OFFLINE
                    MSG_MONITOR_CONFIG_EMPTY
                    MSG_ONLINE_SUCCESS
                    MSG_ONLINE_NEEDSWAP
                    MSG_ONLINE_INCOMPLETE
                    MSG_ONLINE_GOODTOGO
                    MSG_ONLINE_ERROR
                    MSG_ONLINE_UNKNOWN
                    MSG_ONLINE_WAIT
                    MSG_ONLINE_WAIT_LV_SWAP
                    MSG_ONLINE_WAIT_DRACTIVATE
                    MSG_ONLINE_WAIT_TARGET
                    MSG_ONLINE_WAIT_MIRRBACK
                    MSG_ONLINE_WAIT_SWAP
                    MSG_ONLINE_WAIT_RESTORE
                    MSG_ONLINE_GO_AWAY_SOURCE
                    MSG_ONLINE_GO_AWAY_TARGET
                    MSG_ONLINE_GO_AWAY_INVISIBLE
                    MSG_VPAIR_UNKNOWN
                    MSG_VPAIR_RUNNINGGOOD
                    MSG_VPAIR_SOURCEFAILED
                    MSG_VPAIR_DRACTIVATING
                    MSG_VPAIR_DRACTIVATED
                    MSG_VPAIR_RESTORING
                    MSG_VPAIR_NEEDDEACTIVATE
                    MSG_VPAIR_RETRY
                    MSG_VPAIR_RETRY_LIMIT
                    MSG_VPAIR_TASK_RUNNING
                    MSG_VPAIR_TASK_COMPLETE
                    MSG_VPAIR_TASK_RUNNING
                    MSG_VPAIR_TASK_COMPLETE
                    MSG_VPAIR_TASK_FAILED
                    MSG_VPAIR_SWAP_STARTED
                    MSG_VPAIR_SWAP_CANCELED
                    MSG_VPAIR_SWAP_READY
                    MSG_VPAIR_SWAP_WAIT
                    MSG_VPAIR_SWAP_FAILED
                    MSG_VPAIR_RPDEF_SUCCESS
                    MSG_VPAIR_RPDEF_ERR_INVALID
                    MSG_VPAIR_RPDEF_ERR_STATE
                    MSG_VPAIR_RPDEF_ERR_SERVER
                    MSG_VPAIR_RPDEF_ERR_FAILED
                    MSG_VPAIR_VXDISK_MISSING
                    MSG_CONFIG
                    MSG_CONFIG_SUCCESS
                    MSG_CONFIG_FAILURE
                    MSG_CONFIG_INIT
                    MSG_CONFIG_TOPOLOGY
                    MSG_CONFIG_TOPOLOGY_SUCCESS
                    MSG_CONFIG_TOPOLOGY_FAILURE
                    MSG_CONFIG_VOLUMES
                    MSG_CONFIG_VOLUMES_SUCCESS
                    MSG_CONFIG_VOLUMES_FAILURE
                    MSG_CONFIG_ATTR_SUCCESS
                    MSG_CONFIG_ATTR_FAILURE
                    MSG_CONFIG_ATTR_INFO_VOLS
                    MSG_CONFIG_ATTR_INFO_SSNS
                    MSG_CONFIG_ATTR_INFO_SRVS
                    MSG_CONFIG_ERR_PARENTS
                    MSG_CONFIG_ERR_NODELIST
                    MSG_CONFIG_ERR_NODEID
                    MSG_CONFIG_ERR_CLEAR
                    MSG_CONFIG_ERR_VXPORTS
                    MSG_CONFIG_ERR_SCPORTS
                    MSG_CONFIG_ERR_NOPATH
                    MSG_CONFIG_ERR_DISKGROUPS
                    MSG_CONFIG_ERR_VXDISKS
                    MSG_CONFIG_ERR_SERIALS
                    MSG_CONFIG_ERR_LUNS
                    MSG_CONFIG_ERR_RPTS
                    MSG_CONFIG_ERR_RPDEFINE
                    MSG_CONFIG_ERR_LINK
                );
}







sub LOG {

    my $msgID   = shift;
    my $msg     = $MESSAGES{$msgID} || "[message id=$msgID]";
    VCSAG_LOG_MSG(($MESSAGELEVELS{$msgID % 10} || 'I'), $msg . (@_ ? ":\n" . join("\n", map { DUMP($_) } @_) : ''), $msgID);
    map { print $_ "\n\t", ISOTIME, "\n\t", $msg, "\n", (@_ ? join("\n", map { DUMP($_, 2) } @_) . "\n" : '') } @$LOGFILES if $LOGFILES;
    return ($msgID % 10 <= 1)
}







sub TRACE {

    my ($title, $obj, @items) = @_;
    map { print $_ "\n\t", ISOTIME, "\n\t", uc($title), ":\n", DUMP($obj, 2, @items), "\n" } @$LOGFILES if $LOGFILES;
}







sub MSG {

    my $msgID = shift;
    return $MESSAGES{$msgID} || "[message id=$msgID]";
}







sub addLogFiles {

    return unless @_;

    foreach (map { "$_." . strftime('%Y.%m', localtime) . '.log' } grep { not ref($_) } grep { $_ } @_) {

        if (open(my $fh, '>>', $_)) {

            print $fh "\n\n", "*" x 120, "\n", ISOTIME, "\n";
            push @$LOGFILES, $fh;
        }
    }
}





DESTROY {

    map { close($_) } @$LOGFILES;
}



1;