#
# Revision: 3.0.180532
#
# Copyright (c) 2018 Dell Technologies Inc.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# DELL EMC CORPORATION. USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF DELL EMC.
#
package SCTools;
use strict;
use warnings;
use POSIX qw(strftime);




BEGIN {

    use Exporter qw(import);

    our @EXPORT_OK = qw(
                        MIN
                        MAX
                        ISIN
                        GREP
                        FILTER
                        UNIQ
                        GREPUNIQ
                        GREPPROPERTY
                        SAFEPARAMS
                        SAFEVALUE
                        SAFEKEY
                        ISOTIME
                        DATETIME
                        TOSTRING
                        DUMP
                        PRINTDUMP
                        RUN
                        RUN_SINGLE );
}





sub MIN           { my $min = shift; map { $min = $_ if $_ < $min } @_; return $min; }

sub MAX           { my $max = shift; map { $max = $_ if $_ > $max } @_; return $max; }

sub ISIN          { my ($value, @values) = @_; my %filter = map { $_ => 1 } grep { $_ } @values; return exists($filter{$value}) }

sub SAFEPARAMS    { map { /[^-_0-9A-Za-z]/ ? '"' . s/"/\\"/rgi . '"' : $_ } grep { $_ } @_ }

sub SAFEVALUE     { wantarray ? (map { s/[^\w\-\+()|]+/_/rgi } map { $_||'' } @_) : ($_[0]||'') =~ s/[\W]+/_/rgi; }

sub SAFEKEY       { wantarray ? (map { s/[\W]+/_/rgi }         map { $_||'' } @_) : ($_[0]||'') =~ s/[\W]+/_/rgi; }

sub RUN           { my $output = shift; my $cmd = join(' ', @_); @$output = split("\n", `$cmd 2>&1` || ''); return (0 == $?) ? 1 : SCLogger::LOG(SCLogger::MSG_RUN_FAILED(), $cmd, "Error: $?", @$output) }

sub RUN_SINGLE    { my $default = shift; my @output = (); return RUN(\@output, @_) ? $output[0] : $default }

sub ISOTIME       { strftime('%Y-%m-%dT%H:%M:%S', localtime) }

sub DATETIME      { strftime('%Y.%m.%d %H:%M:%S', localtime) }

sub TOSTRING      { my @obj = @_; join('; ', map { _tostring(0, $_) } @obj) }

sub PRINTDUMP     { my ($title, $obj, @items) = @_; print $title, ":\n", DUMP($obj, 1, @items), "\n" }

sub FILTER        { my ($source, $property, @values) = @_; return () unless $source && @values; my %filter = map { $_ => 1 } grep { $_ } @values; return grep { exists($filter{$_->$property}) } grep { ref $_ && $_->can($property) } @$source }

sub UNIQ          { values %{{ map { $_ => $_ } grep { defined $_ } @_ }} }

sub GREPUNIQ      { my ($property, @input) = @_; my %filter = (); grep { not $filter{$_->$property}++ } grep { ref $_ && $_->can($property) } @input }

sub GREPPROPERTY  { my ($property, $value, @input) = @_; grep { $value eq $_->$property } grep { ref($_) && $_->can($property) } @input }

sub GREP          {

    my     ($properties, $values, @input) = @_;
    return () unless $properties && $values && @input;

            $properties = [$properties] unless ref($properties) eq 'ARRAY';
            $values     = [$values]     unless ref($values)     eq 'ARRAY';
    my      %filter     = (map { my $obj = $_; (map { scalar($obj->$_) => $obj } grep { $obj->can($_) } @$properties) } grep {ref} @input);
    return  UNIQ(@filter{@$values});
}

sub DUMP {

    my ($obj, $tab, @items) = @_;

        $tab = $tab || 0;

    unless (ref($obj)) {

        return "\t" x $tab . ($obj || 0);
    }

    elsif (ref($obj) eq 'ARRAY') {

        $obj = { map { $_ => DUMP($obj->[$_], 0, @items) } keys @$obj };
        @items = sort { $a <=> $b } keys %$obj
    }

    elsif (ref($obj) ne 'HASH') {

        return "\t" x $tab . $obj if $obj->isa('SCALAR');
        @items = $obj->properties if (not @items) && $obj->isa('ISerializable');
        $obj = @items ? { map { $obj->can($_) ? ($_ => _tostring(3, $obj->$_)) : ($_ => '(not implemented)') } @items } : { %$obj }
    }

        @items = sort keys %$obj unless @items;
    my  $width = MAX(map {length} @items);

    return join ("\n", map { my @lines = split("\n", _tostring(3, $obj->{$_}));
                            "\t" x $tab . $_ . ' ' x ($width - length $_) . ' : ' . (@lines < 2 ? $lines[0] : join ("\n" . "\t" x $tab . ' ' x ($width+3), @lines) . "\n");
                           } @items);
}





sub _tostring {

    my $lvl = shift || 0;
    my $obj = @_ > 1 ? [@_] : (shift || 0);

    return $obj unless ref($obj);
    return '[' . join('; ', ($lvl > 0) ? map { _tostring($lvl-1, $_) } @$obj : @$obj) . ']' if ref($obj) eq 'ARRAY';
    return($obj->{objectType} || '') . '{' . join('; ', map { "$_: " . ($obj->{$_} || 0)} grep { exists($obj->{$_}) } qw(instanceName instanceId)) . '}' if (ref($obj) eq 'HASH') and (exists($obj->{instanceId}));
    return                             '{' . join('; ', map { "$_: " . ($obj->{$_} || 0)} keys %$obj) . '}'                                              if (ref($obj) eq 'HASH');
    return $obj if $obj->isa('SCALAR');
    return ref($obj) . ' (' . join('; ', (map { join(': ', $_, ($obj->$_ || 0)) } grep { $obj->can($_) } grep { $_ } ($obj->properties))) . ')' if $obj->isa('ISerializable');
    return ref($obj) . ' {' . $obj->name . '}' if $obj->can('name');
    return $obj;
};


1;