#
# Revision: 3.0.180532
#
# Copyright (c) 2018 Dell Technologies Inc.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# DELL EMC CORPORATION. USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF DELL EMC.
#
package sc::api;
use lib qw(. ../);
use strict;
use warnings;
use constant {

    #API object classes
    CLASSObject             => 'sc::object',

    #API constructor input params
    DSMServer               => 'DSMServer',
    DSMServerReplica        => 'DSMServerReplica',
    DSMPort                 => 'DSMPort',
    DSMUser                 => 'DSMUser',
    DSMPass                 => 'DSMPass',

    #API Session Object fields
    APISession              => 'API.Session',
    APISessionID            => 'API.Session.ID',
    APIValidObject          => 'API.IsValid',
    APIUserAgent            => 'API.UserAgent',
    APIVerifySSL            => 'API.VerifySSL',
    APITimeOut              => 'API.TimeOut',
    APIBaseURL              => 'API.BaseURL',
    APIBaseURLFailover      => 'API.BaseURL.Failover',
    APIFailedOver           => 'API.FailedOver.To.DSMReplica',
    APILastResponse         => 'API.LastResponse',
    APICachedFaults         => 'API.CachedFaults',

    #API Response Object fields
    APIResponse             => 'API.Response',
    APIResponseRequest      => 'API.Request',
    APIResponseSuccess      => 'API.Success',
    APIResponseInfo         => 'API.Info',
    APIResponseData         => 'API.Data',
    APIResponseDataRaw      => 'API.Data.Raw',

    #API REST request URLs
    URLLogin                => '/ApiConnection/Login',
    URLLogout               => '/ApiConnection/Logout',
    URLStorageCenters       => '/ApiConnection/ApiConnection/{instanceId}/StorageCenterList',
    URLReplications         => '/StorageCenter/ScReplication',
    URLReplicationsList     => '/StorageCenter/ScReplication/GetList',
    URLRestorePoint         => '/DisasterRecovery/DrScReplicationRestorePoint/{instanceId}',
    URLRestorePoints        => '/DisasterRecovery/DrScReplicationRestorePoint',
    URLRestorePointsList    => '/DisasterRecovery/DrScReplicationRestorePoint/GetList',
    URLReplicationQoS       => '/StorageCenter/ScReplicationQosNode',
    URLReplicationQoSList   => '/StorageCenter/ScReplicationQosNode/GetList',
    URLVolumeInstance       => '/StorageCenter/ScVolume/{instanceId}',
    URLVolumes              => '/StorageCenter/ScVolume',
    URLVolumesList          => '/StorageCenter/ScVolume/GetList',
    URLLiveVolume           => '/StorageCenter/ScLiveVolume/{instanceId}',
    URLLiveVolumes          => '/StorageCenter/ScLiveVolume',
    URLLiveVolumesList      => '/StorageCenter/ScLiveVolume/GetList',
    URLLvSaveRestorePoint   => '/StorageCenter/ScLiveVolume/{instanceId}/SaveRestorePoint',
    URLLvSwapRoles          => '/StorageCenter/ScLiveVolume/{instanceId}/SwapRoles',
    URLDrTasks              => '/DisasterRecovery/DrReplicationAsyncTask',
    URLDrTasksList          => '/DisasterRecovery/DrReplicationAsyncTask/GetList',
    URLDrActivatedVolume    => '/DisasterRecovery/DrScReplicationRestorePoint/{instanceId}/DrActivatedVolume',
    URLDrSettings           => '/DisasterRecovery/DrScReplicationRestorePoint/{instanceId}/DrActivateSetting',
    URLPredefineDR          => '/DisasterRecovery/DrScReplicationRestorePoint/{instanceId}/PredefineDrActivate',
    URLActivateDR           => '/DisasterRecovery/DrScReplicationRestorePoint/{instanceId}/ActivateDr',
    URLDeactivateDR         => '/DisasterRecovery/DrReplicationAsyncTask/{instanceId}/DeactivateDestination',
    URLActivateRestore      => '/DisasterRecovery/DrScReplicationRestorePoint/{instanceId}/Restore',
    URLValidateSingle       => '/DisasterRecovery/DrScReplicationRestorePoint/{instanceId}/ValidateSingle',
    URLValidateAll          => '/DisasterRecovery/DrScReplicationRestorePoint/ValidateAll',
    URLServer               => '/StorageCenter/ScServer/{instanceId}',
    URLServers              => '/StorageCenter/ScServer',
    URLMappingProfiles      => '/StorageCenter/ScMappingProfile',
    URLMappingProfilesList  => '/StorageCenter/ScMappingProfile/GetList',
    URLMappingProfile       => '/StorageCenter/ScMappingProfile/{instanceId}',
    URLMappingProfileModify => '/StorageCenter/ScMappingProfile/{instanceId}/ModifyAdvanced',
    URLHbas                 => '/StorageCenter/ScServerHba',
    URLHbasList             => '/StorageCenter/ScServerHba/GetList',
    URLPreAllocate          => '/StorageCenter/ScVolume/{instanceId}/PreallocateStorage',
    URLVolMappings          => '/StorageCenter/ScVolume/{instanceId}/MappingList',
    URLVolMappingProfiles   => '/StorageCenter/ScVolume/{instanceId}/MappingProfileList',
    URLVolStorageCenter     => '/StorageCenter/ScVolume/{instanceId}/StorageCenter',
    URLRPointLiveVolume     => '/DisasterRecovery/DrScReplicationRestorePoint/{instanceId}/LiveVolume',
    URLSaveRestorePoint     => '/StorageCenter/ScReplication/{instanceId}/SaveRestorePoint',
    URLReplicaRestorePoint  => '/StorageCenter/ScReplication/{instanceId}/RestorePoint',
    URLMapToServer          => '/StorageCenter/ScVolume/{instanceId}/MapToServer',
    URLRemoteSCList         => '/StorageCenter/StorageCenter/{instanceId}/RemoteStorageCenterList',
    URLRemoteSCRemoteSC     => '/StorageCenter/ScRemoteStorageCenter/{instanceId}/RemoteStorageCenter',
    URLTargetReplica        => '/StorageCenter/ScVolume/{instanceId}/ReplicationDestination',
    URLSourceReplicas       => '/StorageCenter/ScVolume/{instanceId}/ReplicationSourceList',
    URLBringPrimaryOnline   => '/StorageCenter/ScLiveVolume/{instanceId}/BringPrimaryOnline',

    #API REST objects attributes
    ATTRInstanceId          => 'instanceId',
    ATTRInstanceName        => 'instanceName',
    ATTRObjectType          => 'objectType',
    ATTRPConnected          => 'connected',
    ATTRApiVersion          => 'apiVersion',
    ATTRHostName            => 'hostName',
    ATTRWebPort             => 'webServicesPort',
    ATTRProvider            => 'provider',
    ATTRProviderVersion     => 'providerVersion',
    ATTRUserName            => 'userName',
    ATTRParent              => 'parent',
    ATTRVolume              => 'volume',
    ATTRTargetVolume        => 'destinationVolume',
    ATTRSourceVolume        => 'sourceVolume',
    ATTRSerialNumber        => 'serialNumber',
    ATTRScSSN               => 'scSerialNumber',
    ATTRSourceSSN           => 'sourceScSerialNumber',
    ATTRTargetSSN           => 'destinationScSerialNumber',
    ATTRSecondarySSN        => 'secondaryScSerialNumber',
    ATTRDestinationAvailable=> 'destinationAvailable',
    ATTRDestVolumeAvailable => 'destinationVolumeAvailable',
    ATTRDeviceID            => 'deviceId',
    ATTRDrActivated         => 'drActivated',
    ATTRDrSettingPreserveLV => 'preserveLiveVolume',
    ATTRDrSettingServers    => 'serverList',
	ATTRDrSettingSyncBefore => 'syncReplicationBeforeActivate',
	ATTRDrSettingActiveSnap => 'useActiveReplay',
	ATTRDrSettingOrigFolder => 'useOriginalVolumesFolder',
    ATTRLiveVolume          => 'liveVolume',
    ATTRLunIdRequested      => 'lunRequested',
    ATTRLunIdUsed           => 'lunUsed',
    ATTRReadOnly            => 'readOnly',
    ATTRConnectivity        => 'connectivity',
    ATTRState               => 'state',
    ATTRStateMessage        => 'stateMessage',
    ATTRStatus              => 'status',
    ATTRStatusMessage       => 'statusMessage',
    ATTRSyncStatus          => 'syncStatus',
    ATTRSyncDataStatus      => 'syncDataStatus',
    ATTRSyncMode            => 'syncMode',
    ATTRSyncType            => 'type',
    ATTRTimeStarted         => 'timeStarted',
    ATTRStepMessage         => 'stepMessage',
    ATTRError               => 'error',
    ATTRErrorCode           => 'errorCode',
    ATTRErrorState          => 'errorState',
    ATTRWarningList         => 'warningList',
    ATTRMethodName          => 'methodName',
    ATTRTaskType            => 'taskType',
    ATTRCurrentStep         => 'currentStep',
    ATTRTotalSteps          => 'totalSteps',
    ATTRRestoreState        => 'restoreState',
    ATTRSuccessful          => 'successful',
    ATTRResult              => 'result',
    ATTRPortWWNs            => 'portWwnList',
    ATTRServer              => 'server',
    ATTRServerHBA           => 'serverHba',
    ATTRReturnValue         => 'returnValue',
    ATTRNeedToDeactivate    => 'needToDeactivate',
    ATTRObjectInstance      => 'objectInstance',
    ATTRReplComplete        => 'replicationComplete',
    ATTRServerList          => 'serverList',
    ATTRSourceSC            => 'sourceStorageCenter',
    ATTRTargetSC            => 'destinationStorageCenter',
    ATTRRemoteStorageCenter => 'remoteStorageCenter',
    ATTRQosNode             => 'qosNode',
    ATTRManagedByLV         => 'managedByLiveVolume',
    ATTRPauseAllowed        => 'pauseAllowed',
    ATTRReplicateActReplay  => 'replicateActiveReplay',
    ATTRToLowestTier        => 'replicateStorageToLowestTier',
    ATTRSimulation          => 'simulation',
    ATTRTransportTypes      => 'transportTypes',
    ATTRTransportTypesSec   => 'secondaryTransportTypes',
    ATTRPrimaryVolume       => 'primaryVolume',
    ATTRPrimaryStatus       => 'primaryStatus',
    ATTRPrimaryRole         => 'primaryRole',
    ATTRSecondaryVolume     => 'secondaryVolume',
    ATTRSecondaryStatus     => 'secondaryStatus',
    ATTRSecondaryRole       => 'secondaryRole',
    ATTRReplicaTarget       => 'replicationDestination',
    ATTRReplicaSource       => 'replicationSource',
    ATTRReplicationState    => 'replicationState',
    ATTRAutoFailover        => 'failoverAutomaticallyEnabled',
    ATTRRepairable          => 'repairable',

    #API Methods params
    PARAM_VOL_DataPageSize                      => 'DataPageSize',
    PARAM_VOL_DataReductionProfile              => 'DataReductionProfile',
    PARAM_VOL_DiskFolder                        => 'DiskFolder',
    PARAM_VOL_GroupQosProfile                   => 'GroupQosProfile',
    PARAM_VOL_ImportToLowestTier                => 'ImportToLowestTier',
    PARAM_VOL_ManagementApplication             => 'ManagementApplication',
    PARAM_VOL_Name                              => 'Name',
    PARAM_VOL_Notes                             => 'Notes',
    PARAM_VOL_ReadCache                         => 'ReadCache',
    PARAM_VOL_Redundancy                        => 'Redundancy',
    PARAM_VOL_ReplayProfileList                 => 'ReplayProfileList',
    PARAM_VOL_Size                              => 'Size',
    PARAM_VOL_StorageCenter                     => 'StorageCenter',
    PARAM_VOL_StorageProfile                    => 'StorageProfile',
    PARAM_VOL_StorageType                       => 'StorageType',
    PARAM_VOL_VolumeFolder                      => 'VolumeFolder',
    PARAM_VOL_VolumeQosProfile                  => 'VolumeQosProfile',
    PARAM_VOL_WriteCache                        => 'WriteCache',
    PARAM_MAP_Advanced                          => 'Advanced',
    PARAM_MAP_AdvancedMapping                   => 'AdvancedMapping',
    PARAM_MAP_ManagementApplication             => 'ManagementApplication',
    PARAM_MAP_Server                            => 'Server',
    PARAM_LV_Dedup                              => 'Dedup',
    PARAM_LV_MinAmountBeforeSwap                => 'MinAmountBeforeSwap',
    PARAM_LV_MinSecondaryPercentBeforeSwap      => 'MinSecondaryPercentBeforeSwap',
    PARAM_LV_MinTimeAsPrimaryBeforeSwap         => 'MinTimeAsPrimaryBeforeSwap',
    PARAM_LV_PrimaryQosNode                     => 'PrimaryQosNode',
    PARAM_LV_PrimaryVolume                      => 'PrimaryVolume',
    PARAM_LV_ReplicateActiveReplay              => 'ReplicateActiveReplay',
    PARAM_LV_ReplicateStorageToLowestTier       => 'ReplicateStorageToLowestTier',
    PARAM_LV_RestoreAutomaticallyEnabled        => 'RestoreAutomaticallyEnabled',
    PARAM_LV_SecondaryAdvancedMapping           => 'SecondaryAdvancedMapping',
    PARAM_LV_SecondaryServer                    => 'SecondaryServer',
    PARAM_LV_SecondaryStorageCenter             => 'SecondaryStorageCenter',
    PARAM_LV_SecondaryVolume                    => 'SecondaryVolume',
    PARAM_LV_SecondaryVolumeAttributes          => 'SecondaryVolumeAttributes',
    PARAM_LV_SecondaryQosNode                   => 'SecondaryQosNode',
    PARAM_LV_StorageCenter                      => 'StorageCenter',
    PARAM_LV_SwapRolesAutomaticallyEnabled      => 'SwapRolesAutomaticallyEnabled',
    PARAM_LV_SyncMode                           => 'SyncMode',
    PARAM_LV_TransportTypes                     => 'TransportTypes',
    PARAM_LV_Type                               => 'Type',
    PARAM_LV_ScSerialNumber                     => 'ScSerialNumber',
    PARAM_RPDR_DrActivateSettings               => 'DrActivateSettings',
    PARAM_RPDR_Name                             => 'Name',
    PARAM_RPDR_PreserveLiveVolume               => 'PreserveLiveVolume',
    PARAM_RPDR_Replay                           => 'Replay',
    PARAM_RPDR_ReplayProfileList                => 'ReplayProfileList',
    PARAM_RPDR_RestorePointInstanceId           => 'RestorePointInstanceId',
    PARAM_RPDR_ServerList                       => 'ServerList',
    PARAM_RPDR_SyncReplicationBeforeActivate    => 'SyncReplicationBeforeActivate',
    PARAM_RPDR_UseActiveReplay                  => 'UseActiveReplay',
    PARAM_RPDR_UseOriginalVolume                => 'UseOriginalVolume',
    PARAM_RPDR_UseOriginalVolumesFolder         => 'UseOriginalVolumesFolder',
    PARAM_RPDR_VolumeFolder                     => 'VolumeFolder',
    PARAM_RPRS_RestoreSettings                  => 'RestoreSettings',
    PARAM_RPRS_MirrorBackOnly                   => 'mirrorBackOnly',
    PARAM_RPRS_DeactivateDestination            => 'automaticallyDeactivateDestination',
    PARAM_RPRS_RecoverLiveVolume                => 'recoverLiveVolume',

    #API REST objects state attribute values
    STATE_UP                => 'Up',
    STATE_DOWN              => 'Down',
    STATE_PARTIAL           => 'Partial',
    STATE_DEGRADED          => 'Degraded',
    STATE_UNKNOWN           => 'Unknown',
    STATE_DRACTIVATED       => 'DrActivated',
    STATE_DRACTIVATING      => 'DrActivating',
    STATE_RUNNINGGOOD       => 'RunningGood',
    STATE_RESTORING         => 'Restoring',
    STATE_NO_REPLICATION    => 'NoReplication',
    STATE_NO_DESTINATION    => 'NoDestinationVolume',
    STATE_NOT_VALIDATED     => 'NotValidated',
    STATE_SOURCE_SC_DOWN    => 'SourceStorageCenterDown',
    STATE_RUNNING           => 'Running',
    STATE_ERROR             => 'Error',
    STATE_FINISHED          => 'Finished',
    STATE_ACTIVATED         => 'Activated',
    STATE_PAUSED            => 'Paused',

    #API other constants used by various API objects
    VALUE_Recovery          => 'Recovery',
    VALUE_Running           => 'Running',
    VALUE_None              => 'None',
    VALUE_ASync             => 'Asynchronous',
    VALUE_Sync              => 'Synchronous',
    VALUE_SyncHA            => 'HighAvailability',
    VALUE_SyncHighConsist   => 'HighConsistency',
    VALUE_Err_StorageCenter => 'StorageCenterError',

    #API Filters
    FILTER_Equals           => 'Equals',
    FILTER_Greater          => 'GreaterThan',

    #API Object Types
    TYPE_StorageCenter      => 'StorageCenter',
    TYPE_Volume             => 'ScVolume',
    TYPE_LiveVolume         => 'ScLiveVolume',
    TYPE_Server             => 'ScPhysicalServer',
    TYPE_ServerCluster      => 'ScServerCluster',
    TYPE_RestorePoint       => 'DrScReplicationRestorePoint',
    TYPE_Task               => 'AsyncTask',
    TYPE_DrTask             => 'DrReplicationAsyncTask',
    TYPE_Replication        => 'ScReplication',
    TYPE_ReplicationQoS     => 'ScReplicationQosNode',
    TYPE_RemoteSC           => 'ScRemoteStorageCenter',
    TYPE_Connection         => 'ApiConnection'
};



use LWP::UserAgent;
use JSON;
use SCLogger;
use connection;
use response;
use storagecenter;
use volume;
use livevolume;
use mappingprofile;
use server;
use hba;
use replication;
use restorepoint;
use drtask;
use remotesc;



my  %APIObjectsFactory = %{ JSON::from_json(<<'API_OBJECTS_DEFINITIONS'
    {
        "StorageCenter"                 :["sc::storagecenter"   , "/StorageCenter/StorageCenter/{instanceId}"                   ],
        "ScVolume"                      :["sc::volume"          , "/StorageCenter/ScVolume/{instanceId}"                        ],
        "ScLiveVolume"                  :["sc::livevolume"      , "/StorageCenter/ScLiveVolume/{instanceId}"                    ],
        "ScReplication"                 :["sc::replication"     , "/StorageCenter/ScReplication/{instanceId}"                   ],
        "ScReplicationQosNode"          :["sc::object"          , "/StorageCenter/ScReplicationQosNode/{instanceId}"            ],
        "DrScReplicationRestorePoint"   :["sc::restorepoint"    , "/DisasterRecovery/DrScReplicationRestorePoint/{instanceId}"  ],
        "DrReplicationAsyncTask"        :["sc::drtask"          , "/DisasterRecovery/DrReplicationAsyncTask/{instanceId}"       ],
        "ScServerCluster"               :["sc::server"          , "/StorageCenter/ScServer/{instanceId}"                        ],
        "ScPhysicalServer"              :["sc::server"          , "/StorageCenter/ScServer/{instanceId}"                        ],
        "ScVirtualServer"               :["sc::server"          , "/StorageCenter/ScServer/{instanceId}"                        ],
        "ScServerHba"                   :["sc::hba"             , "/StorageCenter/ScServerHba/{instanceId}"                     ],
        "ScMapping"                     :["sc::mapping"         , "/StorageCenter/ScMapping/{instanceId}"                       ],
        "ScMappingProfile"              :["sc::mappingprofile"  , "/StorageCenter/ScMappingProfile/{instanceId}"                ],
        "ScRemoteStorageCenter"         :["sc::remotesc"        , "/StorageCenter/ScRemoteStorageCenter/{instanceId}"           ],
        "AsyncTask"                     :["sc::task"            , "/ApiConnection/AsyncTask/{instanceId}"                       ],
        "ApiConnection"                 :["sc::connection"      , "/ApiConnection/ApiConnection"                                ]
    }
API_OBJECTS_DEFINITIONS
    )};



sub OR      { return ( Filter => {FilterType => 'OR',  Filters => [ map { {AttributeName => ($_->[0] || ''), AttributeValue => ($_->[1] || 0), FilterType => ($_->[2] || FILTER_Equals())} } @_ ]} ) }
sub AND     { return ( Filter => {FilterType => 'AND', Filters => [ map { {AttributeName => ($_->[0] || ''), AttributeValue => ($_->[1] || 0), FilterType => ($_->[2] || FILTER_Equals())} } @_ ]} ) }
sub EQUALS  { my $attr = shift; map { [ $attr, $_, FILTER_Equals  ] } @_ }
sub GREATER { my $attr = shift; map { [ $attr, $_, FILTER_Greater ] } @_ }





#@returns sc::api
sub connect {

    my  $class = shift;
    my  $self = bless {@_}, $class;
        $self->{DSMPort()}              = $self->{DSMPort()}         || 3033;
        $self->{APIVerifySSL()}         = $self->{APIVerifySSL()}    || 0;
        $self->{APITimeOut()}           = $self->{APITimeOut()}      || 30;
        $self->{APIBaseURL()}           = $self->{DSMServer()}        ? ('https://'.$self->{DSMServer()}.':'.$self->{DSMPort()}.'/api/rest')        : 'Invalid DSMServer';
        $self->{APIBaseURLFailover()}   = $self->{DSMServerReplica()} ? ('https://'.$self->{DSMServerReplica()}.':'.$self->{DSMPort()}.'/api/rest') : '';
        $self->{APIFailedOver()}        = 0;
        $self->{APISessionID()}         = 0;
        $self->{APISession()}           = 0;
        $self->{APICachedFaults()}      = {};

    my  LWP::UserAgent  $http = $self->{APIUserAgent()} = LWP::UserAgent->new( ssl_opts => { verify_hostname => $self->{APIVerifySSL()}, SSL_verify_mode => $self->{APIVerifySSL()} } );
                        $http->protocols_allowed(['https',]);
                        $http->timeout($self->{APITimeOut()});

    my  HTTP::Headers   $hdrs = $http->default_headers();
                        $hdrs->header('Content-Type', 'application/json');
                        $hdrs->header('x-dell-api-version', '3.1');
                        $hdrs->header('Accept', 'application/json');

    my  HTTP::Request   $reqt = HTTP::Request->new( POST => $self->{APIBaseURL()}.URLLogin());
                        $reqt->headers->authorization_basic($self->{DSMUser()}, $self->{DSMPass()});

    my  HTTP::Response  $resp = $http->simple_request($reqt);
                        $self->{APILastResponse()} = bless(
                               {   APISession()            => $self,
                                   APIResponseRequest()    => $reqt->url(),
                                   APIResponseSuccess()    => $resp->is_success(),
                                   APIResponseInfo()       => $resp->status_line,
                                   APIResponseDataRaw()    => $resp->decoded_content,
                                   APIResponseData()       => []
                               }, 'sc::response' );

    if ($resp->is_success()) {

                $hdrs->header('cookie', $resp->header('set-cookie'));
                $self->{APIFailedOver()}    = 0;
                $self->{APISessionID()}     = $resp->header('set-cookie');
                $self->{APISession()}       = $self->inflate({ATTRObjectType() => 'ApiConnection'});

        TRACE  MSG(MSG_CONNECTION_PRIMARY), $self->{APISession()};
        return  $self;
    }

    else {

        TRACE MSG(MSG_CONNECTION_PRIMARY_ERR), $resp;

        if ($self->{APIBaseURLFailover()}) {

            $reqt = HTTP::Request->new( POST => $self->{APIBaseURLFailover()}.URLLogin());
            $reqt->headers->authorization_basic($self->{DSMUser()}, $self->{DSMPass()});
            $resp = $http->simple_request($reqt);

            if ($resp->is_success()) {

                        $hdrs->header('cookie', $resp->header('set-cookie'));
                        $self->{APIFailedOver()}    = 1;
                        $self->{APIBaseURL()}       = $self->{APIBaseURLFailover()};
                        $self->{APISessionID()}     = $resp->header('set-cookie');
                        $self->{APISession()}       = $self->inflate({ATTRObjectType() => 'ApiConnection'});

                TRACE MSG(MSG_CONNECTION_FAILOVER), $self->{APISession()};
                return  $self;
            }

            else {

                TRACE MSG(MSG_CONNECTION_FAILOVER_ERR), $resp;
            }
        }
    }

    return 0;
};





#@method
#@returns sc::response
sub _submit {

    my ($self, $method, $url, $id, $async, @body) = @_;
                        $url =~ s|\{instanceId\}|$id|gi if defined $id;
                        $url = $self->{APIBaseURL()}.$url;

    my  sc::response    $cache = $self->_get_cached_faults($id);
                        return $cache if $cache;

    my  HTTP::Request   $rqt = HTTP::Request->new($method => $url);
                        $rqt->content(JSON::to_json({@body})) if @body;
                        $rqt->headers->header('async' => 'true') if $async;

    my  HTTP::Response  $rsp = $self->{APIUserAgent()}->simple_request($rqt);

    my  sc::response    $result = bless(
                                    {   APISession()            => $self,
                                        APIResponseRequest()    => $url,
                                        APIResponseSuccess()    => $rsp->is_success,
                                        APIResponseInfo()       => $rsp->status_line,
                                        APIResponseDataRaw()    => $rsp->decoded_content,
                                        APIResponseData()       => []
                                    }, 'sc::response' );

                        $self->{APILastResponse()} = $result;

    if ($result->success) {

        if (my $data = JSON::from_json( $result->data || '{}' )) {

            $result->{APIResponseData()} = ('ARRAY' eq ref $data) ? $data : [$data];
        }
    }

    else {

        TRACE '[API] REQUEST FAILED', $result;
        $self->_set_cached_faults($id, $result) if $id && $rsp->code == 404;
    }

    return $result;
};





#@method
sub _set_cached_faults {

    my $self = shift;
    my $id   = shift;
    my $resp = shift;
    return unless $id && ref $resp;
    return unless $resp->{APIResponseDataRaw()};

    my $data = JSON::from_json($resp->{APIResponseDataRaw()});
    return unless $data;
    return unless 'HASH' eq ref $data;
    return unless $data->{ATTRErrorCode()} eq VALUE_Err_StorageCenter();

    my $ssn  = $1 if $id =~ /\A(\d+).(\d+)\z/;
       $self->{APICachedFaults()}->{$id}  = $resp;
       $self->{APICachedFaults()}->{$ssn} = $resp if $ssn;
}





#@method
#@returns sc::response
sub _get_cached_faults {

    my $self = shift;
    my $id   = shift;
    return 0 unless $id;

    my $ssn  = $1 if $id =~ /\A(\d+).(\d+)\z/;
    my $resp = $self->{APICachedFaults()}->{$id} || $ssn && $self->{APICachedFaults()}->{$ssn};
    return 0 unless $resp;

    TRACE '[API] CACHED FAULT RESPONSE', [{id => $id, ssn => $ssn}, $resp];
    return $resp;
}





#@method
#@returns sc::object
sub wrap {

    my  $self   = shift;
    my  $proto  = shift || {}; #$prototype should be a hash with ObjectType and InstanceId fields;
    my  $target = shift || '';
    my  $type   = $proto->{ATTRObjectType()} || '';
    my  $valid  = $type && exists($APIObjectsFactory{$type});
    my  $class  = $valid ? $APIObjectsFactory{$type}->[0] : $target;

        $proto->{APIValidObject()} = $valid && exists $proto->{ATTRInstanceId()};
        $proto->{ATTRInstanceId()} = 0  unless exists $proto->{ATTRInstanceId()};
        $proto->{ATTRObjectType()} = '' unless exists $proto->{ATTRObjectType()};
        $proto->{APISession()}     = $self;
        $proto = bless($proto, $class || CLASSObject());

    if ($target and not $proto->isa($target)) {

        $proto->{APIValidObject()} = 0;
        $proto = bless($proto, $target);
    }

    return $proto;
};





#@method
#@returns sc::object
sub inflate {

    my  $self   = shift;
    my  $proto  = shift || {}; #$prototype should be a hash with ObjectType and InstanceId fields;
    my  $class  = shift || '';
    my  $method = shift || 'GET';
    my  $type   = $proto->{ATTRObjectType()} || 0;
    my  $id     = $proto->{ATTRInstanceId()} || 0;
    my  $url    = $type ? ($APIObjectsFactory{$type}->[1] || '') : '';

    if ($url) {

        my  $rsp = $self->_submit($method, $url, $id, 0);

        if ($rsp->success) {

            my sc::object   $obj = $rsp->asObject($class);
                            $obj->{APIValidObject()} &&= ($obj->type eq $type);
                            $obj->{APIValidObject()} &&= ($obj->id eq $id) if $id;
            return $obj;
        }

        return $self->wrap({%$rsp, ATTRObjectType() => $type}, $class);
    }

    TRACE '[API] FAILURE TO CONVERT PROTOTYPE INTO AN API OBJECT', $proto;
    return bless({ APISession() => $self, APIValidObject() => 0 }, ($class || CLASSObject()));
};







#@method
sub url             { shift->{APIBaseURL()}     || ''}


#@method
sub id              { shift->{APISessionID()}   || 0 }


#@method
sub isFailedOver    { shift->{APIFailedOver()}  || 0 }

#@method
#@returns sc::response
sub lastResponse    { shift->{APILastResponse()}|| bless({}, 'sc::response') }

#@method
#@returns sc::connection
sub connection      { my $self = shift; $self->{APISession()} || $self->dummy_object(TYPE_Connection) }


#@method
sub disconnect      { my $self = shift; TRACE("DISCONNECT", $self, qw(url id)); $self->{APISessionID()} = 0 if $self->_submit('POST', URLLogout())->success }


#@method
DESTROY             { my $self = shift; $self->{APIUserAgent()}->simple_request(HTTP::Request->new(POST => $self->{APIBaseURL()}.URLLogout())) if $self->{APIUserAgent()} && $self->{APIBaseURL()} && $self->{APISessionID()} }

#@method
#@returns sc::response
sub GET             { my ($self, $url, @body) = @_; return $self->_submit('GET',    $url, 0, 0, @body) };

#@method
#@returns sc::response
sub PUT             { my ($self, $url, @body) = @_; return $self->_submit('PUT',    $url, 0, 0, @body) };

#@method
#@returns sc::response
sub DELETE          { my ($self, $url, @body) = @_; return $self->_submit('DELETE', $url, 0, 0, @body) };

#@method
#@returns sc::response
sub POST            { my ($self, $url, @body) = @_; return $self->_submit('POST',   $url, 0, 0, @body) };

#@method
#@returns sc::response
sub POST_ASYNC      { my ($self, $url, @body) = @_; return $self->_submit('POST',   $url, 0, 1, @body) };







#@method
#@returns sc::object
sub dummy_object    { my ($self, $type, %fields) = @_; return $self->wrap({ %fields, ATTRObjectType() => $type }) }

#@method
#@returns sc::object
sub get_object      { my ($self, $type, $id, $class) = @_; return $self->inflate({ATTRObjectType() => $type, ATTRInstanceId() => $id}, $class) }

#@method
#@returns sc::object
sub select_objects  { my ($self, $url, $param, @values) = @_; return $self->_submit('POST', $url, 0, 0, OR(EQUALS($param, @values)))->asObject }







#@method
#@returns sc::storagecenter
sub arrays          {   my     $self= shift;
                        return $self->GET (URLStorageCenters)->asObject('sc::storagecenter') }


#@method
#@returns sc::server
sub servers         {   my     $self= shift;
                        return $self->GET (URLServers)->asObject('sc::server') }


#@method
#@returns sc::restorepoint
sub restorePoints   {   my     $self= shift;
                        return $self->GET (URLRestorePoints)->asObject('sc::restorepoint') unless @_;
                        return $self->POST(URLRestorePointsList, OR(EQUALS(ATTRSourceSSN, @_), EQUALS(ATTRTargetSSN, @_)))->asObject('sc::restorepoint') }

#@method
#@returns sc::replication
sub replications    {   my     $self= shift;
                        return $self->GET (URLReplications)->asObject('sc::replication') unless @_;
                        return $self->POST(URLReplicationsList(), OR(EQUALS(ATTRScSSN, @_), EQUALS(ATTRTargetSSN, @_)))->asObject('sc::replication') }

#@method
#@returns sc::livevolume
sub livevolumes     {   my     $self= shift;
                        return $self->GET (URLLiveVolumes)->asObject('sc::livevolume') unless @_;
                        return $self->POST(URLLiveVolumesList(), OR(EQUALS(ATTRScSSN, @_), EQUALS(ATTRSecondarySSN, @_)))->asObject('sc::livevolume') }

#@method
#@returns sc::drtask
sub drTasks         {   my     $self= shift;
                        return $self->GET (URLDrTasks)->asObject('sc::drtask') unless @_;
                        return $self->POST(URLDrTasksList, OR(EQUALS(ATTRState, @_)))->asObject('sc::drtask') }

#@method
#@returns sc::volume
sub volumes         {   my     $self= shift;
                        return $self->GET (URLVolumes)->asObject('sc::volume') unless @_;
                        return $self->POST(URLVolumesList, OR(EQUALS(ATTRSerialNumber, @_)))->asObject('sc::volume') }

#@method
#@returns sc::hba
sub hbas            {   my     $self= shift;
                        return $self->GET (URLHbas)->asObject('sc::hba') unless @_;
                        return $self->POST(URLHbasList, OR(EQUALS(ATTRScSSN, @_)))->asObject('sc::hba') }







#@method
#@returns sc::volume
sub get_volume      {   my ($self, $id) = @_; return $self->get_object(TYPE_Volume(),           $id, 'sc::volume') }

#@method
#@returns sc::restorepoint
sub get_rpoint      {   my ($self, $id) = @_; return $self->get_object(TYPE_RestorePoint(),     $id, 'sc::restorepoint') }

#@method
#@returns sc::storagecenter
sub get_sc          {   my ($self, $id) = @_; return $self->get_object(TYPE_StorageCenter(),    $id, 'sc::storagecenter') }

#@method
#@returns sc::task
sub get_task        {   my ($self, $id) = @_; return $self->get_object(TYPE_Task(),             $id, 'sc::task') }

#@method
#@returns sc::server
sub get_server      {   my ($self, $id) = @_; return $self->get_object(TYPE_Server(),           $id, 'sc::server') }






#@method
#@returns sc::response
sub validateAll     {   shift->POST_ASYNC(URLValidateAll)->asObject()   }


#@method
#@returns sc::restorepoint
sub select_rpoints  {   my ($self, $param, @values) = @_;
                        return $self->POST(URLRestorePointsList, OR(EQUALS($param, @values)))->asObject('sc::restorepoint') }



1;