#
# Revision: 3.0.180532
#
# Copyright (c) 2018 Dell Technologies Inc.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# DELL EMC CORPORATION. USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF DELL EMC.
#
package sc::restorepoint;
use strict;
use warnings;
use parent 'sc::object';
use drtask;


use constant {

    STATE_00_Unknown            =>  0,
    STATE_00_Error              =>  1,
    STATE_00_NoReplication      =>  2,
    STATE_01_RunningGood        => 10,
    STATE_02_SourceFailed       => 20,
    STATE_03_FailoverInProgress => 30,
    STATE_04_TargetActivated    => 40,
    STATE_05_RestoreInProgress  => 50,
    STATE_06_NeedToDeactivate   => 60
};



our %RPSTATUS = (

    sc::api::STATE_UNKNOWN()    => 0,
    sc::api::STATE_DOWN()       => 1,
    sc::api::STATE_DEGRADED()   => 2,
    sc::api::STATE_UP()         => 3
);



#@method
#@override
sub properties          { qw (sourceVolumeName id sourceSSN sourceVolumeID targetSSN targetVolumeID state status syncStatus isDRActivated isTargetAvailable isRunningGood isLiveVolume) }



#@method
sub status              { $RPSTATUS{$_[0]->{sc::api::ATTRStatus()}||''} || $RPSTATUS{sc::api::STATE_UNKNOWN()} }


#@method
#@returns sc::restorepoint
sub validate            { my ($self) = @_; $self->POST(sc::api::URLValidateSingle())->success ? $self->api->get_rpoint($self->id) : $self }


#@method
#@returns sc::response
sub predefine           { my ($self, $settings) = @_; $self->POST(sc::api::URLPredefineDR(),
                                                                  sc::api::PARAM_RPDR_DrActivateSettings() => ($settings||{})) }


#@method
#@returns sc::drtask
sub activateDR          { my ($self, $settings) = @_; $self->POST(sc::api::URLActivateDR(), Test => 'false', $settings ?
                                                                 (sc::api::PARAM_RPDR_DrActivateSettings() => $settings) : () )->asObject('sc::drtask') }


#@method
#@returns sc::drtask
sub activateRestore     { my ($self, %settings) = @_; $self->POST(sc::api::URLActivateRestore(),
                                                                  sc::api::PARAM_RPRS_RestoreSettings() => {
                                                                  sc::api::PARAM_RPRS_DeactivateDestination() => 1,
                                                                  sc::api::PARAM_RPRS_RecoverLiveVolume() => 1, %settings} )->asObject('sc::drtask') }
#@method
#@returns sc::drtask
sub activateMirrorBack  { my ($self, %settings) = @_; $self->POST(sc::api::URLActivateRestore(),
                                                                  sc::api::PARAM_RPRS_RestoreSettings() => {
                                                                  sc::api::PARAM_RPRS_MirrorBackOnly() => 1,
                                                                  sc::api::PARAM_RPRS_RecoverLiveVolume() => 1, %settings} )->asObject('sc::drtask') }
#@method
#@returns sc::response
sub delete              { $_[0]->DELETE(sc::api::URLRestorePoint) }



#@method
sub settings            { $_[0]->GET(sc::api::URLDrSettings())->asHash }


#@method
#@returns sc::livevolume
sub liveVolume          { $_[0]->GET(sc::api::URLRPointLiveVolume())->asObject('sc::livevolume') }


#@method
#@returns sc::volume
sub activatedVolume     { $_[0]->GET(sc::api::URLDrActivatedVolume())->asObject('sc::volume') }


#@method
#@returns sc::storagecenter
sub targetSC            { $_[0]->api->inflate($_[0]->{sc::api::ATTRTargetSC()}, 'sc::storagecenter') }


#@method
#@returns sc::volume
sub targetVolume        { $_[0]->api->inflate($_[0]->{sc::api::ATTRTargetVolume()}, 'sc::volume') }


#@method
#@returns sc::volume
sub sourceVolume        { $_[0]->api->inflate($_[0]->{sc::api::ATTRSourceVolume()}, 'sc::volume') }



#@method
sub targetVolumeName    { $_[0]->{sc::api::ATTRTargetVolume()}->{sc::api::ATTRInstanceName()} || '' }



#@method
sub sourceVolumeName    { $_[0]->{sc::api::ATTRSourceVolume()}->{sc::api::ATTRInstanceName()} || '' }



#@method
sub targetVolumeID      { $_[0]->{sc::api::ATTRTargetVolume()}->{sc::api::ATTRInstanceId()} || 0 }



#@method
sub sourceVolumeID      { $_[0]->{sc::api::ATTRSourceVolume()}->{sc::api::ATTRInstanceId()} || 0 }



#@method
sub activatedVolumeID   { $_[0]->isDRActivated ? $_[0]->activatedVolume->id : 0 }



#@method
sub sourceSSN           { $_[0]->{sc::api::ATTRSourceSSN()} || 0 }



#@method
sub targetSSN           { $_[0]->{sc::api::ATTRTargetSSN()} || 0 }



#@method
sub syncStatus          { $_[0]->{sc::api::ATTRSyncDataStatus()} || 0 }



#@method
sub state               { $_[0]->{sc::api::ATTRState()} || sc::api::STATE_UNKNOWN }



#@method
sub isTargetAvailable   { $_[0]->{sc::api::ATTRDestVolumeAvailable()} && $_[0]->{sc::api::ATTRDestinationAvailable()} }



#@method
sub isSourceFailed      { $_[0]->isValid && ($_[0]->state eq sc::api::STATE_SOURCE_SC_DOWN()) }



#@method
sub isDRActivated       { $_[0]->isValid && ($_[0]->state eq sc::api::STATE_DRACTIVATED()) && $_[0]->{sc::api::ATTRDrActivated()} }



#@method
sub isRestoring         { $_[0]->isValid && ($_[0]->state eq sc::api::STATE_RESTORING())   && $_[0]->{sc::api::ATTRDrActivated()} }



#@method
sub isRunningGood       { $_[0]->isValid && ($_[0]->state eq sc::api::STATE_RUNNINGGOOD()) && ($_[0]->status >= $RPSTATUS{sc::api::STATE_UP()}) }



#@method
sub isLiveVolume        { $_[0]->isValid && $_[0]->{sc::api::ATTRLiveVolume()} }



#@method
sub isNeedValidation    { $_[0]->isValid && $_[0]->state eq sc::api::STATE_NOT_VALIDATED() }



#@method
sub isErrorState        {

    my $self = shift;
    my %states = map { $_ => 1} (sc::api::STATE_DRACTIVATING(), sc::api::STATE_DRACTIVATED(), sc::api::STATE_RESTORING(), sc::api::STATE_RUNNINGGOOD(), sc::api::STATE_NOT_VALIDATED(), sc::api::STATE_SOURCE_SC_DOWN());
    return $self->isValid && not exists($states{$self->state});
}



#@method
sub isDefinedState      {

    my $self = shift;
    my %states = map { $_ => 1} (STATE_01_RunningGood, STATE_04_TargetActivated, STATE_02_SourceFailed);
    return $self->isValid && exists($states{$self->classify});
}



#@method
sub drTasks             {

    my  $self   = shift;
    my  @tasks  = scalar(@_)
                ? $self->POST(sc::api::URLDrTasksList(), sc::api::OR(sc::api::EQUALS(sc::api::ATTRState(), @_)))->asObject('sc::drtask')
                : $self->GET(sc::api::URLDrTasks())->asObject('sc::drtask');

    return grep { my sc::drtask $t = $_; $t->isValid && ($t->subjectID eq $self->id) } @tasks;
}



#@method
sub classify            {

    my $self = shift;
    return STATE_00_Unknown()               if !(defined $self);
    return STATE_00_Error()                 if !($self->isValid);
    return STATE_00_Error()                 if  ($self->state eq sc::api::STATE_NO_REPLICATION());
    return STATE_00_Error()                 if  ($self->state eq sc::api::STATE_NO_DESTINATION());
    return STATE_03_FailoverInProgress()    if  ($self->state eq sc::api::STATE_DRACTIVATING());
    return STATE_05_RestoreInProgress()     if  ($self->state eq sc::api::STATE_RESTORING());
    return STATE_04_TargetActivated()       if  ($self->state eq sc::api::STATE_DRACTIVATED() && $self->status >= $RPSTATUS{sc::api::STATE_DEGRADED()});
    return STATE_01_RunningGood()           if  ($self->state eq sc::api::STATE_RUNNINGGOOD() && $self->status >= $RPSTATUS{sc::api::STATE_UP()});
    return STATE_02_SourceFailed()          if  ($self->state eq sc::api::STATE_SOURCE_SC_DOWN());
    return STATE_02_SourceFailed()          if !($self->isDRActivated || $self->sourceVolume->isUp);
    return STATE_00_Unknown();
}



1;



#control                    : DrScReplicationControl{instanceName: DR SC Pair 48956 - 60689; instanceId: 48956.60689}
#destinationAvailable       : 1
#destinationScSerialNumber  : 60689
#destinationStorageCenter   : StorageCenter{instanceName: Storage Center 60689; instanceId: 60689}
#destinationVolume          : ScVolume{instanceName: vcs_02; instanceId: 60689.1565}
#destinationVolumeAvailable : 1
#drActivated                : 1
#instanceId                 : 48956.60689.701
#instanceName               : RestorePoint 48956.60689.701
#lastSaveTime               : 2017-11-30T07:52:32Z
#lastSyncTime               : 2017-11-30T07:53:27Z
#lastValidateTime           : 2017-12-01T15:27:10Z
#liveVolume                 : 1
#objectType                 : DrScReplicationRestorePoint
#recommendSyncDataCurrent   : 0
#sourceScSerialNumber       : 48956
#sourceStorageCenter        : StorageCenter{instanceName: Storage Center 48956; instanceId: 48956}
#sourceVolume               : ScVolume{instanceName: vcs_02; instanceId: 48956.886}
#state                      : DrActivated
#status                     : Degraded
#syncDataStatus             : Current
#syncMode                   : HighAvailability



#API.Session                : sc::api=HASH(0x1cd5f58)
#control                    : DrScReplicationControl{instanceName: DR SC Pair 48956 - 60689; instanceId: 48956.60689}
#destinationAvailable       : 1
#destinationScSerialNumber  : 60689
#destinationStorageCenter   : StorageCenter{instanceName: Storage Center 60689; instanceId: 60689}
#destinationVolume          : ScVolume{instanceName: Repl of A-Saffron-Quorum; instanceId: 60689.1584}
#destinationVolumeAvailable : 1
#drActivated                : 0
#instanceId                 : 48956.60689.706
#instanceName               : RestorePoint 48956.60689.706
#lastSaveTime               : 2017-12-01T15:25:04Z
#lastSyncTime               : 2017-12-01T15:27:10Z
#lastValidateTime           : 2017-12-01T15:27:10Z
#liveVolume                 : 1
#objectType                 : DrScReplicationRestorePoint
#recommendSyncDataCurrent   : 0
#sourceScSerialNumber       : 48956
#sourceStorageCenter        : StorageCenter{instanceName: Storage Center 48956; instanceId: 48956}
#sourceVolume               : ScVolume{instanceName: A-Saffron-Quorum; instanceId: 48956.905}
#state                      : RunningGood
#status                     : Up
#syncDataStatus             : Current
#syncMode                   : HighAvailability



#API.Session                : sc::api=HASH(0x1be6330)
#control                    : DrScReplicationControl{instanceName: DR SC Pair 60689 - 48956; instanceId: 60689.48956}
#destinationAvailable       : 1
#destinationScSerialNumber  : 48956
#destinationStorageCenter   : StorageCenter{instanceName: Storage Center 48956; instanceId: 48956}
#destinationVolume          : ScVolume{instanceName: R2_LVA_09; instanceId: 48956.522}
#destinationVolumeAvailable : 0
#drActivated                : 0
#instanceId                 : 60689.48956.446
#instanceName               : RestorePoint 60689.48956.446
#lastSaveTime               : 2017-11-30T06:25:04Z
#lastSyncTime               : 2017-11-30T06:28:01Z
#lastValidateTime           : 2017-12-01T15:26:42Z
#liveVolume                 : 1
#objectType                 : DrScReplicationRestorePoint
#recommendSyncDataCurrent   : 0
#sourceScSerialNumber       : 60689
#sourceStorageCenter        : StorageCenter{instanceName: Storage Center 60689; instanceId: 60689}
#sourceVolume               : ScVolume{instanceName: LVA_09; instanceId: 60689.1033}
#state                      : NoSourceOrDestinationVolumes
#status                     : Down
#syncDataStatus             : Current
#syncMode                   : HighAvailability