#
# Revision: 3.0.180532
#
# Copyright (c) 2018 Dell Technologies Inc.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# DELL EMC CORPORATION. USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF DELL EMC.
#
package sc::volume;
use strict;
use warnings;
use parent 'sc::object';
use lib qw(. ../);
use SCTools qw(FILTER GREPPROPERTY);





#@method
#@override
sub properties      { qw (name serial id status isUp isLiveVolume isSource isTarget) }



#@method
sub serial          { $_[0]->{sc::api::ATTRSerialNumber()}  || 0 };



#@method
sub deviceId        { $_[0]->{sc::api::ATTRDeviceID()}      || 0 };



#@method
sub ssn             { $_[0]->{sc::api::ATTRScSSN()}|| 0 };



#@method
sub status          { $_[0]->{sc::api::ATTRStatus()}        || sc::api::STATE_UNKNOWN() }



#@method
sub isUp            { $_[0]->isValid && ($_[0]->status eq sc::api::STATE_UP()) }



#@method
sub isLiveVolume    { $_[0]->isValid && ($_[0]->{sc::api::ATTRLiveVolume()} || 0) }



#@method
sub isTarget        { $_[0]->isValid && ($_[0]->{sc::api::ATTRReplicaTarget()} || 0) }



#@method
sub isSource        { $_[0]->isValid && ($_[0]->{sc::api::ATTRReplicaSource()} || 0) }


#@method
#@returns sc::response
sub delete          { $_[0]->DELETE_ASYNC(sc::api::URLVolumeInstance()) }


#@method
#@returns sc::task
sub preallocate     { $_[0]->POST_ASYNC(sc::api::URLPreAllocate())->asObject('sc::task') }


#@method
#@returns sc::storagecenter
sub sc              { $_[0]->GET(sc::api::URLVolStorageCenter())->asObject('sc::storagecenter') }


#@method
sub mappings        { grep { !$_->isSC } $_[0]->GET(sc::api::URLVolMappingProfiles())->asObject('sc::mappingprofile') }


#@method
#@returns sc::mappingprofile
sub mapToServer     { my ($self, $server) = @_;
                          $self->POST (sc::api::URLMapToServer(),
                                            sc::api::PARAM_MAP_Server() => ref($server) ? $server->id : $server)->asObject('sc::mappingprofile') }

#@method
#@returns sc::replication
sub targetReplication { $_[0]->GET(sc::api::URLTargetReplica())->asObject('sc::replication') }



#@method
#@returns sc::replication
sub livevolume      {

    my  $self = shift;

    if ($self->isLiveVolume) {

        my @lvs = $self->api->livevolumes;

        foreach (GREPPROPERTY('primaryID', $self->id, @lvs)) {

            return $_ if $_->isValid;
        }

        foreach (GREPPROPERTY('secondaryID', $self->id, @lvs)) {

            return $_ if $_->isValid;
        }
    }

    return $self->api->dummy_object(sc::api::TYPE_LiveVolume);
}



#@method
#@returns sc::replication
sub replications    {

    my $self = shift;
    return () unless $self->isTarget or $self->isSource;

    my @replicas = ();
    push @replicas, $self->GET(sc::api::URLTargetReplica())->asObject('sc::replication') if $self->isTarget;
    push @replicas, $self->GET(sc::api::URLSourceReplicas())->asObject('sc::replication') if $self->isSource;
    return grep {$_->isValid} grep {ref} @replicas;
}


#@method
#@returns sc::task
sub newLiveVolume      {

    my  $self       = shift; return $self->api->dummy_object(sc::api::TYPE_Task()) unless ($self->isValid);
    my  $targetSSN  = shift || $self->sc->remoteNodes->remoteSSN;
        $targetSSN  = $targetSSN->id if ref($targetSSN) && $targetSSN->isa('sc::storagecenter');
    my  $localQoS   = shift || $self->api->select_objects(sc::api::URLReplicationQoSList(), sc::api::ATTRScSSN(), $self->ssn)->id;
    my  $targetQoS  = shift || $self->api->select_objects(sc::api::URLReplicationQoSList(), sc::api::ATTRScSSN(), $targetSSN)->id;
        $localQoS   = $localQoS ->id if ref($localQoS)  && $localQoS ->isa('sc::object');
        $targetQoS  = $targetQoS->id if ref($targetQoS) && $targetQoS->isa('sc::object');

    return $self->api->dummy_object(sc::api::TYPE_Task()) if (ref($targetSSN) || ref($localQoS) || ref($targetQoS));
    return $self->api->dummy_object(sc::api::TYPE_Task()) unless ($targetSSN && $localQoS && $targetQoS);
    return $self->POST (sc::api::URLLiveVolumes(),
                        sc::api::PARAM_LV_StorageCenter()                   => $self->ssn,
                        sc::api::PARAM_LV_PrimaryVolume()                   => $self->id,
                        sc::api::PARAM_LV_PrimaryQosNode()                  => ref($localQoS)  ? $localQoS->id  : $localQoS,
                        sc::api::PARAM_LV_SecondaryQosNode()                => ref($targetQoS) ? $targetQoS->id : $targetQoS,
                        sc::api::PARAM_LV_SecondaryStorageCenter()          => $targetSSN,
                        sc::api::PARAM_LV_SwapRolesAutomaticallyEnabled()   => 1,
                        sc::api::PARAM_LV_ReplicateActiveReplay()           => 1,
                        sc::api::PARAM_LV_ReplicateStorageToLowestTier()    => 0,
                        sc::api::PARAM_LV_SyncMode()                        => sc::api::VALUE_SyncHA(),
                        sc::api::PARAM_LV_Type()                            => sc::api::VALUE_Sync(),
                        sc::api::PARAM_LV_SecondaryVolumeAttributes()       => { sc::api::PARAM_VOL_Name() => 'R2_'.$self->name }
    )->asObject();

    #sc::api::PARAM_LV_SecondaryVolume()                 => 'SecondaryVolume',
    #sc::api::PARAM_LV_SecondaryAdvancedMapping()       => {},
    #sc::api::PARAM_LV_SecondaryServer()                => 'SecondaryServer'
}


1;


	#active                    : 0
	#cmmDestination            : 0
	#cmmSource                 : 0
	#configuredSize            : 357556027392 Bytes
	#deleteAllowed             : 1
	#deviceId                  : 6000d31000bf3c000000000000000077
	#flashOptimized            : 1
	#hostCacheEnabled          : 0
	#inRecycleBin              : 0
	#instanceId                : 48956.108
	#instanceName              : vcs_volume_03
	#liveVolume                : 0
	#mapped                    : 1
	#migration                 : 0
	#name                      : vcs_volume_03
	#objectType                : ScVolume
	#portableVolumeDestination : 0
	#replayAllowed             : 0
	#replicationDestination    : 0
	#replicationSource         : 1
	#scName                    : Storage Center 48956
	#scSerialNumber            : 48956
	#secureDataUsed            : 0
	#serialNumber              : 0000bf3c-00000077
	#status                    : Down
	#statusMessage             : 0
	#storageType               : ScStorageType{instanceName: Assigned - Redundant - 512 KB; instanceId: 48956.1}
	#volumeFolder              : ScVolumeFolder{instanceName: VERITAS; instanceId: 48956.3}
	#volumeFolderPath          : VERITAS/
	#vpdId                     : 119